class Document:
    def __init__(self, mod, node, env):
        self.mod = mod
        self.env = env

        self.kindbrackets = mod.kindbrackets
        self.eitherbrackets = mod.eitherbrackets
        self.anykind = mod.anykind

        self.out = []
        self.localstack = []
        self.outstack = []
        self.output_directives = []
        self.document_title = None
        self.specified_definitions = None
        self.macro_args = None
        self.subdoc = None
        self.in_in = 0
        self.macro_stack = []

        node.accept(self)

        self.result = self.node_of_taci('string', '', self.out, 0)

    @property
    def doc_name_node(self):
        return self.node_of_taci('text', self.get_doc_name())

    def _visit_children(self, node):
        E = self.mod.ReportedError
        for ch in node.children:
            try:
                ch.accept(self)
            except E:
                pass

    def _visit_subjects(self, subjects):
        for s in subjects:
            self.out.append(self.node_of_taci('symbol', s.tgtfullname))

    def _visit_gets(self, node, what):
        self._visit_subjects(
            getattr(self.get_arg_subject(node), 'get_%s' % what)())

    def ap_text(self, text):
        self.out.append(self.node_of_taci('text', text, (), 0))

    def close(self, chktag=None, n=1, chk=None, get=False):
        for i in range(n):
            out, tag, arg = self.outstack.pop()
            node = self.node_of_taci(tag, arg, self.out)
            if not get:
                out.append(node)
            self.out = out
            if chk is not None:
                assert chk is out
        if chktag is not None:
            assert chktag == tag
        return node

    def error(self, msg, context=None, **kwds):
        msg = 'Document: ' + msg
        more = [('Macro call site.', ms[0]) for ms in self.macro_stack]
        more.reverse()

        self.env.error(msg, context, more=more, **kwds)

    def error_no_sub_element(self, node, subelement):
        self.error('No such subelement allowed in the enclosing element %r.' %
                   node.tag, subelement, exception=None)

    def expand_arg(self, arg):
        i = len(self.localstack) - 1
        while i >= 0:
            argnode = self.localstack[i].get(arg)
            if argnode is not None:
                assert argnode.tag == 'symbol'
                return argnode.arg.strip()
            i -= 1
        return arg

    def expand_list(self, li):
        oldout = self.out
        self.out = newout = []
        for node in li:
            node.accept(self)
        self.out = oldout
        return newout

    def expand_node(self, node, optarg=0, optmore=0):
        arg, rest = self.get_arg_rest(
            node, optarg=optarg, optmore=optmore, nostrip=1)
        return self.node_of_taci(node.tag, arg, rest, node.index, node.src)

    def gen_char(self, char):
        self.gen_tag('char', char)

    def gen_document_header(self):
        self.open('document_header')
        self.close()

    def gen_document_trailer(self):
        self.open('document_trailer')
        self.close()

    def gen_link_to(self, s, text=None, children=()):
        if text is not None:
            children = list(children)
            children.append(self.node_of_taci('text', text))
        ln = self.get_link_name(s)
        self.gen_tag('link_to', ln, children)

    def gen_localname(self, s):
        self.gen_link_to(s, s.get_local_name())

    def gen_symbol(self, arg):
        self.out.append(self.node_of_taci('symbol', arg))

    def gen_tag(self, tag, arg='', children=()):
        self.out.append(self.node_of_taci(tag, arg, children))

    def gen_text(self, text):
        self.gen_tag('text', text)

    def get_arg_only(self, node):
        arg, rest = self.get_arg_rest(node)
        if rest:
            self.error('Node has extra children, only 1 arg or child expected')
        return arg

    def get_arg_rest(self, node, optarg=0, optmore=0, nostrip=0):
        arg, rest = node.get_arg_rest(nostrip=nostrip)
        rest = self.expand_list(rest)
        if arg:
            arg = self.expand_arg(arg)
        else:
            if not (rest and rest[0].tag == 'symbol'):
                if not optarg:
                    self.error(
                        'Argument on line or as next children expected.', node)
            else:
                arg = rest[0].arg.strip()
                rest = rest[1:]
        if rest and rest[0].tag == 'symbol':
            if not optmore:
                self.error('More arguments than expected.', rest[0])
        return arg, rest

    def get_arg_subject(self, node):
        arg = self.get_arg_only(node)
        return self.get_subject(arg, node)

    def get_arg_subjects_rest(self, node):
        args, rest = self.get_arglist_rest(node)
        return [self.get_subject(a, node) for a in args], rest

    def get_arglist_only(self, node):
        args, rest = self.get_arglist_rest(node)
        if rest:
            self.error_no_sub_element(node, rest[0])
        return args

    def get_arglist_rest(self, node):
        args = []
        for arg in node.get_arglist():
            if not arg:
                self.error("Empty argument in arg list", node)
            arg = self.expand_arg(arg)
            args.append(arg)
        rest = []
        for a in self.expand_list(node.children):
            if a.tag == 'symbol':
                if rest:
                    self.error(
                        'Argument elements must be first in subelements.',
                        a,
                        exception=None)
                args.append(a.arg.strip())
            else:
                rest.append(a)
        return args, rest

    def get_cur_subject(self, node):
        sd = self.subdoc
        if not sd:
            self.error('No subject defined in current environment.', node)
        return sd.subject

    def get_doc_name(self):
        return self.document_name

    def get_filers(self, output_dir):
        if not self.output_directives:
            print('Document %r: No output directives' % self.name)
        filers = []
        r = self.get_result()
        name = self.get_doc_name()
        for (handler, opts) in self.output_directives:
            print('processing', handler, opts, name)
            filers.append(handler.doc2filer(
                self, r, name, output_dir, opts, self.mod.IO))
        return filers

    def get_link_name(self, a):
        return a.get_link_name()

    def get_macro_args(self, node):
        args = self.macro_args
        if args is None:
            self.error('Is not in macro', node)
        return args

    def get_result(self):
        return self.result

    def get_subject(self, name, node=None):
        return self.env.get_descr_by_name(name, node)

    def node_of_taci(self, *args):
        return self.mod.node_of_taci(*args)

    def open(self, tag, arg=''):
        self.outstack.append((self.out, tag, arg))
        self.out = []
        return self.out

    def subdoc_do(self, m, f):
        sd = SubDoc(self, m)
        osd = self.subdoc
        self.subdoc = sd
        sd.subdoc = sd
        try:
            f(sd)
        finally:
            self.subdoc = osd

    def visit_args(self, node):
        if self.macro_args is None:
            self.error('Not in macro', node)
        else:
            names = self.get_arglist_only(node)
            if len(names) != len(self.macro_args):
                self.error('%d args passed, here is %d names' % (len(self.macro_args), len(names)),
                           node)
            self.localstack.append(dict([(x.strip(), self.macro_args[i])
                                         for i, x in enumerate(names)]))

    def visit_arguments_of(self, node):
        self._visit_gets(node, 'arguments')

    def visit_attr_name_of(self, node):
        self.gen_symbol(self.get_arg_subject(node).get_attr_name())

    def visit_attributes_of(self, node):
        self._visit_gets(node, 'attributes')

    def visit_block(self, node):
        self.out.append(self.expand_node(node, optarg=1))

    def visit_default(self, node):
        self.out.append(self.expand_node(node, optarg=1, optmore=1))

    def visit_define(self, node):
        arg, rest = self.get_arg_rest(node)

        arg = self.get_link_name(self.get_subject(arg, node))
        self.out.append(self.node_of_taci(
            node.tag, arg, rest, node.index, node.src))

    def visit_defines(self, node):
        sd = self.specified_definitions
        if sd is None:
            sd = self.specified_definitions = []
        sd.extend(self.get_arglist_only(node))

    def visit_description_of(self, node):
        self.get_arg_subject(node).gen_description_doc(self)

    def visit_document(self, node):
        self.document_name = node.arg.strip()
        self.open('document')
        self._visit_children(node)
        if self.document_title is None:
            self.open('document_title')
            self.gen_text('GSL document %s' % self.document_name)
            self.close()
        self.close()

    def visit_document_title(self, node):
        self.document_title = node
        self.out.append(self.document_title)

    def visit_for(self, node):
        varname = node.get_namearg()

        if not node.children:
            self.error('For loop without subelements.', node)

        if not (node.children[0].tag == 'in'):
            self.error("First subelement of for loop must be 'in'.",
                       node.children[0])
        inode = node.children[0]

        names = self.get_arglist_only(inode)

        body = node.children[1:]
        if not body:
            self.error('For loop without body.', node)
        for name in names:
            self.localstack.append({
                varname: self.node_of_taci(
                    'symbol',
                    name, (), node.index)})
            try:
                for ch in body:
                    ch.accept(self)
            finally:
                self.localstack.pop()

    def visit_gsml(self, node):
        arg, rest = node.get_arg_rest()
        if arg:
            rest = [self.mod.node_of_taci(
                'text', arg, (), node.index, node.src)]+list(rest)
        self.open('block')
        for a in rest:
            if a.tag == 'text':
                a = self.mod.node_of_gsml(a.arg.strip())
            a.accept(self)
        self.close()

    def visit_id_of(self, node):
        self.ap_text(self.get_arg_subject(node).get_id_name())

    def visit_in(self, node):
        self.in_in += 1
        self.visit_default(node)
        self.in_in -= 1

    def visit_kind_of(self, node):
        self.gen_symbol(self.get_arg_subject(node).get_kind_name())

    def visit_label(self, node):
        subject = self.get_cur_subject(node)
        arg, rest = self.get_arg_rest(node)
        name = subject.get_link_name() + '.label:'+arg
        self.open('define', name)
        self.close()
        for r in rest:
            self.out.append(r)

    def visit_link_to(self, node):
        arg, rest = self.get_arg_rest(node)
        self.gen_link_to(self.get_subject(arg, node), children=rest)

    def visit_man_page_of(self, node):
        self.open('to_document_only')
        self.open('man_page_mode')
        subjects, rest = self.get_arg_subjects_rest(node)
        if rest:
            self.error_no_sub_element(node, rest[0])
        for subject in subjects:
            self.subdoc_do(subject, lambda sd: sd.gen_man_page(subject))

        self.close()
        self.close()

    def visit_mappings_of(self, node):
        self._visit_gets(node, 'mappings')

    def visit_meta(self, node):
        arg = node.arg.strip()
        if arg:
            colon = arg.find(':')
            if colon <= 0:
                self.error('Argument to meta, if any,  must be of the form <name>:<content>.',
                           node)
            name = arg[:colon].strip()
            content = arg[colon+1:].strip()
            mknode = self.mod.node_of_taci
            ch = (mknode('name', name), mknode(
                'content', content)) + node.children
            node = mknode('meta', '', ch)
        self.out.append(node)

    def visit_name_of(self, node):
        self.gen_text(self.get_arg_subject(node).get_name())

    def visit_output(self, node):
        mode, rest = self.get_arg_rest(node)
        modes = [x.strip() for x in mode.split(',')]
        for mode in modes:
            try:
                handler_name = self.mod.output_handlers[mode.lower()]
            except KeyError:
                self.error('Unknown output mode: %r. Expected one of %r.' % (
                    mode, list(self.mod.output_handlers.keys())),
                    node,
                    exception=None)
            else:
                handler = getattr(self.mod, handler_name)
                self.output_directives.append((handler, rest))

    def visit_ref(self, node):
        self.gen_text(' ')
        subject = self.get_cur_subject(node)
        arg, rest = self.get_arg_rest(node)
        text = arg
        if arg.startswith('.'):
            dl = arg.find('.', 1)
            if dl < 0:
                dl = len(arg)
            tag = arg[1:dl].strip()
            name = arg[dl+1:].strip()
        else:
            tag = 'myfile'
            name = arg
        if tag == 'mykind':
            idn = subject.get_link_name()
            if name:
                idn = idn + '.' + name
                text = name
            else:
                text = idn.split('.')[-1]
        elif tag == 'myfile':
            idn = subject.get_link_name()
            idn = '.'.join(idn.split('.')[:2])
            if name:
                idn = idn + '.' + name
                text = name
        else:
            self.error('Invalid tag: %r in reference.' % tag, node)

        if not rest:
            rest = [self.node_of_taci('text', text)]
        self.out.append(self.node_of_taci(
            'link_to', idn, rest, node.index))

    def visit_specified_definitions(self, node):
        if node.arg.strip() or node.children:
            self.error('No argument or subelement allowed for element %r.' % node.tag, node,
                       exception=None)
        if self.specified_definitions is None:
            self.error('No definitions have been specified.', node)
        for s in self.specified_definitions:
            self.out.append(self.node_of_taci(
                'symbol', s, (), node.index, node.src))

    def visit_symbol(self, node):
        arg = self.get_arg_only(node)
        if arg != node.arg.strip():
            node = self.node_of_taci(node.tag, arg, (), node.index, node.src)
        self.out.append(node)

    def visit_synopsis_of(self, node):
        self.open('to_document_only')
        self.open('man_page_mode')
        m = self.get_arg_subject(node)
        self.subdoc_do(m, lambda sd: sd.gen_mapping_doc(m))
        self.close()
        self.close()

    def visit_test_of(self, node):
        args, rest = self.get_arg_subjects_rest(node)
        for kind in args:
            self.open('to_tester_only')
            self.out.append(self.node_of_taci(
                node.tag, kind, rest, node.index, node.src))
            self.close()

    def visit_take_all(self, node):
        for a in self.get_macro_args(node):
            self.out.append(a)

    def visit_take_first(self, node):
        args = self.get_macro_args(node)
        if not args:
            self.error('No argument passed', node)
        self.out.append(args[0])

    def visit_take_rest(self, node):
        args = self.get_macro_args(node)
        if not args:
            self.error('No argument passed', node)
        for ch in args[1:]:
            self.out.append(ch)

    def visit_text(self, node):
        self.out.append(node)

    def visit_use(self, node):
        macrocolonarg, args = self.get_arg_rest(node)
        colonpos = macrocolonarg.find(':')
        if colonpos <= 0:
            macroname = macrocolonarg
        else:
            macroname = macrocolonarg[:colonpos].strip()
            macroarg = macrocolonarg[colonpos+1:].strip()
            if not macroarg:
                self.error('Argument must be of form <macroname> or <macroname>:<macroarg>.',
                           node)
            macroarg = self.expand_arg(macroarg)
            args = [self.node_of_taci('symbol', macroarg)] + args
        macro = self.get_subject(macroname, node)

        o = (self.localstack, self.macro_args)
        try:
            self.macro_stack.append([node])
            self.localstack = []
            self.macro_args = args
            self._visit_children(macro.use(args))
        finally:
            (self.localstack, self.macro_args) = o
            self.macro_stack.pop()


class Attributes:
    d_tag = 'attributes'

    def __init__(self, as_):
        self.as_ = as_

    def find_kind_aspects(self):
        return self.as_[0].find_kind_aspects()

    def get_link_name(self):
        return self.as_[0].mod.tgt_prefix+'(%s)' % ','.join([x.get_link_name() for x in self.as_])

    def get_name(self):
        return ', '.join([x.get_name() for x in self.as_])

    def get_kind(self):
        return self.as_[0].get_kind()

    def get_self_name(self):
        return self.as_[0].get_self_name()

    def find_aspects(self, tag):
        return self.as_[0].find_aspects(tag)

    def is_method(self):
        self.as_[0].is_method()


class SubDoc(Document):
    def __init__(self, parent, subject):
        self.__dict__.update(parent.__dict__)
        self.parent = parent
        self.subject = subject
        self.level = 0
        self.no_ret = 0
        self.use_self = None

    def combine_attrs_of_same_kind(self, kas):
        if len(kas) <= 1:
            return kas
        nkas = []
        attrs = []
        for ka in kas:
            t = ka.d_tag
            if t != 'attribute':
                nkas.append(ka)
                continue
            for (i, as_) in attrs:
                a = as_[0]
                if (a.src.node is ka.src.node
                        and len(a.aspects) == len(ka.aspects)):
                    as_.append(ka)
                    break
            else:
                attrs.append((len(nkas), [ka]))
                nkas.append(ka)
        for (i, as_) in attrs:
            if len(as_) > 1:
                nkas[i] = Attributes(as_)
        return nkas

    def combine_attrs_of_same_kind_and_description(self, kas):
        return self.combine_attrs_of_same_kind(kas)

    def gen_anything(self):
        self.open('strong')
        self.gen_text(' ' + self.anykind)
        self.close()

    def gen_argref(self, a):
        # a : kind
        # a = kind
        t = a.d_tag
        if t == 'arg':
            self.gen_posarg_name(a)
            self.gen_colon()
        elif t == 'key_arg':
            self.gen_keyarg_name(a)
            self.gen_assign()
        else:
            assert 0
        self.gen_ref(a.get_kind())

    def gen_arguments(self, args):
        def flatten(args):
            f = []
            for a in args:
                if a.d_tag in ('args', 'seq'):
                    f.extend(flatten(a.find_arg_aspects()))
                else:
                    f.append(a)
            return f

        def gen_sycomma():
            if sycomma:
                self.gen_text(sycomma[0])
            sycomma[:] = [', ']

        def clr_sycomma():
            sycomma[:] = []

        def gen_lbracket(b):
            if sycomma:
                self.gen_text(' ')
            self.gen_text(b)
            clr_sycomma()

        def gen_rbracket(b):
            self.gen_text(b)
            sycomma[:] = [' ']

        def gen_su(text, sup='sup'):
            self.open(sup)
            self.open('strong')
            self.gen_text(text)
            self.close()
            self.close()

        def gen_taggy(tag, args, func, brackets='[]'):
            su = 'sup'
            colon = ':'
            if tag:
                self.gen_text(' ')
                gen_su(tag+colon, su)
                sycomma[:] = []
            if len(args) != 1:
                gen_lbracket(brackets[0])
                func(args)
                gen_rbracket(brackets[1])
            else:
                clr_sycomma()
                func(args)

        def gen_or(asp, sep,
                   orparneed=False,   # Set to True if sequences needs parentheses between or
                   sup=1
                   ):

            if asp:
                if len(asp) == 1:
                    gen_arg(asp[0])
                    return
                gen_arg(asp[0], parneed=orparneed)
                for ch in asp[1:]:
                    if sup:
                        self.open('sup')
                        self.open('strong')
                    if callable(sep):
                        sep()
                    else:
                        self.gen_text(sep)
                    if sup:
                        self.close()
                        self.close()
                    clr_sycomma()
                    gen_arg(ch, parneed=orparneed)

        def gen_arg(a, parneed=0):
            t = a.d_tag
            if t in ('arg', 'key_arg'):
                gen_sycomma()
                self.gen_argref(a)
            elif t == 'alt':
                args = a.find_arg_aspects()
                gen_taggy('alt', args, lambda args: gen_or(args, ' or '))
            elif t == 'no_arg':
                self.gen_text('[]')
            elif t == 'draw':
                args = a.find_arg_aspects()
                if len(args) <= 1:
                    gen_lbracket(' [')
                    gen_arg(args[0])
                    gen_rbracket(']')
                else:
                    gen_taggy('draw', args, lambda args: gen_or(
                        args, ' , ', sup=0))
            elif t == 'optionals':
                args = a.find_arg_aspects()
                for s in args:
                    gen_lbracket(' [')
                    gen_arg(s)
                gen_rbracket(']'*len(args))

            elif t == 'superkind':
                gen_sycomma()
                self.gen_localname(a)
            elif t in ('seq', ):
                args = a.find_arg_aspects()
                gen_taggy('seq', args, lambda args: gen_or(args, ' , ', sup=0))
            elif t in ('args', 'seq'):
                gen_args(a.find_arg_aspects(), parneed)
            elif t == 'repeat':
                gen_taggy(a.get_arg(), a.find_arg_aspects(), gen_args)

            else:
                assert 0

        def gen_args(args, parneed=0):
            args = flatten(args)

            def ga(args):
                for a in args:
                    gen_arg(a)
            if parneed and len(args) > 1:
                #gen_taggy('', args, ga, brackets='<>')
                gen_taggy('1', args, ga)
            else:
                ga(args)
        sycomma = []
        gen_args(args)

    def gen_assign(self):
        self.open('strong')
        self.gen_char('nbsp')
        self.gen_text('=')
        self.gen_char('nbsp')
        self.close()

    def gen_attribute_def(self, a):
        def gen_dt(do_kind=0):
            if dt_done:
                return
            dt_done.append(1)
            self.open('dd')
            if not define_done:
                link_name = self.get_link_name(a)
                self.open('define', link_name)
            if s:
                self.out.append(s)
            if s:
                self.open('code')
                self.gen_text('.')
                self.close()
            self.open('strong')
            self.open('big')
            self.gen_attribute_name(a)
            self.close()
            self.close()
            if not define_done:
                self.close()
                define_done.append(1)

            if do_kind or not kas:
                if len(kas) == 1 and kas[0].d_tag == 'mapping':
                    self.gen_mapping_kind(kas[0], 1)
                    self.open('dl')
                    self.gen_mapping_description(kas[0])
                    self.close()
                else:
                    self.gen_colon()
                    self.gen_def(a.get_kind())
                kind_done.append(1)
            self.close('dd')

        def gen_afterkind(a):
            dt_done.pop()
            gen_dt(1)
            kind_done.append(1)

        define_done = []
        dt_done = []
        kind_done = []
        kas = a.find_kind_aspects()
        s = self.get_self_node(a)

        self.level += 1
        for d in a.find_aspects('*'):
            t = d.d_tag
            if t == 'description':
                gen_dt(0)
                self.open('dd')
                self.open('dl')
                self.gen_description_def(d)
                self.close()
                self.close()
                continue
            if d in kas:
                if dt_done or kind_done:
                    pass
                if not dt_done:
                    gen_dt(do_kind=1)
                elif not kind_done:
                    gen_afterkind(a)
            elif t == 'self':
                pass
            else:
                assert 0

        if not dt_done:
            gen_dt(do_kind=1)
        self.level -= 1

    def gen_attribute_name(self, a):
        self.gen_name(a)

    def gen_attribute_ref(self, a):

        s = self.get_self_node(a)
        if s:
            self.out.append(s)
        self.open('big')
        if s:
            self.open('code')
            self.gen_text('.')
            self.close()

        link_name = self.get_link_name(a)
        self.open('link_to', link_name)
        self.gen_attribute_name(a)
        self.close()
        self.close()

        kas = a.find_kind_aspects()

        if len(kas) == 1 and kas[0].d_tag == 'mapping':
            self.gen_mapping_kind(kas[0])
        else:
            self.gen_colon()
            self.gen_ref(a.get_kind())

    def gen_attributes_def(self, a):
        self.gen_attribute_def(a)

    def gen_attributes_ref(self, a):
        self.gen_attribute_ref(a)

    def gen_colon(self):
        self.open('spc_colonkind')
        self.close()

    def gen_comment_def(self, d):
        pass

    def gen_comment_ref(self, d):
        pass

    def gen_condition_def(self, cond):
        self.open('dt')
        self.gen_condition_ref(cond, 1)
        self.close()
        self.level += 1

        for d in cond.find_aspects('*'):
            t = d.d_tag
            if t == 'description':
                self.gen_description_dd(d)
            elif t == 'python_code':
                self.open('dd')
                self.open('dl')
                self.open('dt')
                self.open('strong')
                self.gen_text('Python code: ')
                self.close()
                self.open('code')
                self.gen_text(d.src.node.arg.strip())
                self.close()
                self.close()
                ctx = d.find_aspects('in context')
                if ctx:
                    self.open('dd')
                    self.open('dl')
                    for ct in ctx:
                        self.open('dt')
                        self.open('strong')
                        self.gen_text('in context: ')
                        self.close()
                        self.open('code')
                        self.gen_text(ct.src.node.arg.strip())
                        self.close()
                        self.close()
                    self.close()
                    self.close()
                self.close('dl')
                self.close('dd')

        self.level -= 1

    def gen_condition_ref(self, cond, define=0):
        link_name = self.get_link_name(cond)
        if define:
            self.open('define', link_name)
        else:
            self.open('link_to', link_name)
        self.open('strong')
        self.open('big')
        self.gen_text(cond.get_def_name())
        self.close()
        self.close()
        self.close()  # define

        self.gen_text('(')
        self.gen_text(', '.join(cond.get_arg_names()))
        self.gen_text(')')

    def gen_constructor_def(self, c):
        self.open('define', self.get_link_name(c))
        self.close()
        for cc in c.args:
            self.open('dt')
            self.gen_link_to(cc)
            self.close()

        for d in c.find_aspects('description'):
            self.gen_description_dd(d)

    def gen_constructor_ref(self, c):
        self.gen_self(c)
        self.gen_text(' = ')
        self.gen_ref(c.args[0])

    def gen_constructor_descriptions(self, li):
        self.gen_constructor_syn(li, 1)

    def gen_constructor_syn(self, li, desc=0):

        ccs = []
        descs = []
        cdccs = []
        for c in li:
            ds = c.find_aspects('description')
            descs.extend(ds)
            ccs.extend(c.args)
            cdccs.append((c, ds, c.args))

        if desc and not descs:
            return

        self.open('dt')
        if desc:
            self.open('h2')
        else:
            self.open('big')
            self.open('strong')
        if descs:
            if desc:
                self.open('define', li[0].get_link_name())
            else:
                self.open('link_to', li[0].get_link_name())
        hd = 'Constructors'
        if len(ccs) == 1:
            hd = hd.rstrip('s')
        self.gen_text(hd)
        if descs:
            self.close()

        self.close()
        if not desc:
            self.close()
        self.close('dt')
        self.open('dd')
        self.open('dl')

        for c, ds, ccs in cdccs:
            for cc in ccs:

                self.open('block')
                self.gen_ref(cc.env)
                conselfnode = self.close(get=1)

                self.open('dt')

                self.use_self = conselfnode
                self.gen_ref(cc)
                self.use_self = None

                self.close('dt')

            if desc:
                for d in ds:
                    self.gen_description_dd(d)
        self.close('dl')
        self.close('dd')

    def gen_def(self, a):
        getattr(self, 'gen_%s_def' % a.d_tag)(a)

    def gen_delitem_def(self, op):
        self.open('dt')
        self.gen_delitem_ref(op, 1)
        self.close()
        self.gen_mapping_description(op)

    def gen_delitem_ref(self, op, subdescript=0):
        link_name = self.get_link_name(op)

        if subdescript:
            self.open('define', link_name)
            self.open('strong')
            self.open('big')
            self.gen_text('del ')
            self.close()
            self.close()
            self.close()
        else:
            self.open('link_to', link_name)
            self.gen_text('del ')
            self.close()

        self.gen_self(op)
        args = op.get_arguments()
        self.gen_text('[')
        self.gen_arguments(args)
        self.gen_text(']')

    def gen_description(self, k):
        ats = k.find_aspects('*')
        ats = self.combine_attrs_of_same_kind_and_description(ats)
        self.gen_descriptions(ats)

    def gen_description_dd(self, d):
        self.open('dd')
        d.gen_doc(self)
        self.close()

    def gen_description_def(self, d):
        self.gen_description_dd(d)

    def gen_description_descriptions(self, li):
        self.gen_outer_dt('Description')
        for d in li:
            self.gen_description_dd(d)

    def gen_description_ref(self, d):
        pass

    def gen_description_syn(self, li):
        pass

    def gen_descriptions(self, ats, use_attr_header=1):
        if not ats:
            return
        tab = self.sortup_aspects(ats)

        for typ, li in tab:
            try:
                try:
                    gen_desc = getattr(self, 'gen_%s_descriptions' % typ)
                except AttributeError:
                    hd = typ
                    if (len(li) > 1):
                        hd = hd + 's'
                    hd = hd.capitalize().replace('_', ' ')
                    self.gen_outer_dt(hd)
                    for a in li:
                        self.gen_def(a)
                else:
                    gen_desc(li)
            except self.mod.ReportedError:
                pass

    def gen_either_def(self, k):
        self.gen_either_ref(k)

    def gen_either_ref(self, k):
        self.open('strong')
        self.open('sup')
        self.gen_text(' either:')
        self.close()
        self.close()
        self.gen_text(self.eitherbrackets[0])
        kas = k.get_alt_kinds()
        self.gen_ref(kas[0])
        for ka in kas[1:]:
            self.open('strong')
            self.open('sup')
            self.gen_text(' or ')
            self.close()
            self.close()
            self.gen_ref(ka)
        self.gen_text(self.eitherbrackets[1])

    def gen_example_descriptions(self, egs):
        e = 'Example'
        if len(egs) > 1:
            e += 's'
        self.gen_outer_dt(e)
        for eg in egs:
            self.open('dd')
            self.open('pre')
            ct = eg.get_ctx_text()
            if ct:
                if not ct.endswith('\n'):
                    ct += '\n'
                self.gen_text(ct)
            et = eg.get_ex_text()
            self.open('strong')
            self.gen_text('return ')
            self.gen_text(eg.get_ex_text())
            self.close()
            self.close()
            self.close()
            continue

            self.open('dd')
            self.open('code')
            self.gen_text(eg.get_ex_text())
            ct = eg.get_ctx_text()
            self.close()

            if ct:
                self.open('em')
                self.gen_text(' # in context:')
                self.close()

            self.close()
            if ct:
                if '\n' in ct:
                    self.open('pre')
                    self.gen_text(ct)
                    self.close()
                else:
                    self.open('dd')
                    self.open('code')
                    self.gen_text(ct)
                    self.close()
                    self.close()

                return

                self.open('dd')
                self.open('dl')
                self.open('dt')
                self.open('strong')
                self.gen_text('In context')
                self.close()
                self.close()
                self.open('dd')
                if '\n' in ct:
                    self.open('pre')
                else:
                    self.open('code')
                self.gen_text(ct)
                self.close()
                self.close()
                self.close()
                self.close('dd')

    def gen_example_syn(self, eg):
        pass

    def gen_function_operator_def(self, op):
        self.open('dd')
        self.gen_function_operator_ref(op, 1)
        self.open('dl')
        self.gen_mapping_description(op)
        self.close()
        self.close()

    def gen_function_operator_ref(self, op, subdescript=0):
        link_name = self.get_link_name(op)

        if not subdescript:
            self.open('link_to', link_name)
        else:
            self.open('define', link_name)
        self.open('big')
        self.open('strong')
        self.gen_text(op.src.node.arg.strip())
        self.close()
        self.close()
        self.close()
        self.gen_text('(')
        self.gen_self(op)

        for a in op.get_arguments():
            t = a.d_tag
            if t == 'arg':
                self.gen_argref(a)
            else:
                assert 0

        self.gen_text(')')

        self.gen_returns(op, subdescript)

    def gen_header(self, m):
        link_name = self.get_link_name(m)
        self.open('define', link_name)
        self.open('h1')
        self.gen_text(m.get_name())
        self.close()
        self.close()

    def gen_inplace_operator_def(self, op):
        self.gen_operator_def(op)

    def gen_inplace_operator_ref(self, op, subdescript=0):
        self.gen_operator_ref(op, subdescript)

    def gen_keyarg_name(self, a):
        self.open('code')
        self.gen_name(a)
        self.close()

    def gen_kind_aspects(self, ats, defi):
        if not ats:
            self.gen_anything()
            return
        self.gen_text(self.kindbrackets[0])
        self.open('dl')
        for a in ats:
            if a.d_tag in ('kind',) and not a.is_synthetic:
                self.open('dd')
                self.open('em')
                self.gen_text('Subkind of: ')
                self.close()
                if defi:
                    self.gen_def(a)
                else:
                    self.gen_ref(a)
                self.close()
            else:
                if defi:
                    self.open('dd')
                    self.gen_def(a)
                    self.close()
                else:
                    self.open('dd')
                    self.gen_ref(a)
                    self.close()
        self.open('dd')
        self.gen_text(self.kindbrackets[1])
        self.close()
        self.close()

    def gen_kind_def(self, k):
        kas = k.find_kind_aspects()
        self.gen_kind_refodef(k, 1)

    def gen_kind_of_def(self, k):
        self.gen_kind_of_ref(k)

    def gen_kind_of_ref(self, k):
        kas = k.find_kind_aspects()
        if len(kas) == 1:
            self.gen_ref(kas[0])
        else:
            assert 0  # to be tested

    def gen_kind_ref(self, k, defi=0):
        self.gen_kind_refodef(k, 0)

    def gen_kind_refodef(self, k, defi=0):
        if not k.is_synthetic:
            self.gen_localname(k)
            return

        kas = k.find_kind_aspects()
        kas = self.combine_attrs_of_same_kind(kas)
        self.gen_kind_aspects(kas, defi)

    def gen_man_page(self, m):
        self.gen_header(m)
        self.open('dl')
        self.gen_Name(m)
        self.gen_synopsis(m)
        self.gen_description(m)
        self.close()

    def gen_mapping_def(self, a):
        self.gen_mapping_tag(a, 1)
        self.gen_mapping_kind(a, 1, 1)

        self.open('dl')

        self.gen_mapping_description(a)

        self.close()

    def gen_mapping_description(self, m):
        def find_named_args(m):
            na = []
            asp = m.find_arg_aspects()
            for a in asp:
                t = a.d_tag
                if t in ('arg', 'key_arg'):
                    na.append(a)
                else:
                    na.extend(find_named_args(a))
            return na

        def gen_arguments_descriptions(m):
            na = find_named_args(m)
            if not na:
                return
            namedesc = {}
            ada = []
            for a in na:
                t = a.d_tag
                if t in ('arg', 'key_arg'):
                    da = a.find_aspects('description')
                    ada.append((a, da))
                    if da:
                        namedesc[(t, a.get_name())] = 1
                else:
                    assert 0

            if namedesc:
                label = 'Argument'
                if len(namedesc) > 1:
                    label += 's'
                self.gen_outer_dt(label)
                self.open('dd')
                self.open('dl')
                for a, da in ada:
                    t = a.d_tag
                    if not da and (t, a.get_name()) in namedesc:
                        # This arg is considered to be described elsewhere, Notes Aug 10 2005.
                        # This is a bit sublte, may do for now...
                        continue
                    self.open('dt')
                    self.gen_argref(a)
                    self.close()
                    for d in da:
                        self.gen_description_dd(d)
                self.close()
                self.close('dd')

        def gen_condition_ref(a):
            if a.is_not:
                self.gen_text('not ')

            d = a.get_definition()

            if d is None:
                self.gen_text(a.cond_doc_name)
            else:
                self.open('link_to', self.get_link_name(d))
                self.gen_text(a.cond_doc_name)
                self.close()

            if a.arg_names:
                self.gen_text('(')
                # self.open('var')
                comma = 0
                for an in a.arg_names:
                    if comma:
                        self.gen_text(', ')
                    comma = 1
                    if an.startswith('<') and an.endswith('>'):
                        self.open('em')
                        self.gen_text(an[1:-1])
                        self.close()
                    else:
                        # I think it normally is clearer to not have
                        # slanted argument names
                        self.gen_text(an)
                # self.close()
                self.gen_text(')')

        def gen_condition_desc(a):
            ds = a.find_aspects('description')
            for d in ds:
                self.gen_description_dd(d)

        def gen_conditions_description(m, asp='precondition'):
            self.open('dd')
            self.open('dl')
            pres = m.find_aspects(asp)
            if pres:
                self.open('dt')
                self.open('strong')
                hd = asp.capitalize()
                if len(pres) > 1:
                    hd = hd + 's'
                self.gen_text(hd)
                self.close()
                self.close()
                self.open('dd')
                self.open('dl')
                for pre in pres:
                    self.open('dt')
                    gen_condition_ref(pre)
                    self.close()
                    gen_condition_desc(pre)
                self.close()
                self.close()
            self.close()
            self.close()

        def gen_description(m):
            asp = m.find_aspects()
            args_described = 0
            pre_described = 0
            post_described = 0
            last_t = None
            last_h = None
            for a in asp:
                t = a.d_tag
                if t == 'description':
                    self.gen_description_dd(a)
                elif t == 'returns':
                    rds = a.find_aspects('description')
                    if rds:
                        self.open('dd')
                        self.open('dl')

                        self.open('dt')
                        self.open('strong')
                        self.gen_text('Returns ')
                        self.close()
                        rds[0].gen_doc(self)
                        rds = rds[1:]
                        self.close('dt')

                        for rd in rds:
                            self.gen_description_dd(rd)
                        self.close('dl')
                        self.close('dd')

                elif t in ('precondition', ):
                    if not pre_described:
                        gen_conditions_description(m, t)
                        pre_described = 1
                elif t in ('postcondition',):
                    if not post_described:
                        gen_conditions_description(m, t)
                        post_described = 1

                elif t in ('equation',):
                    self.open('dd')
                    self.open('dl')
                    self.open('dt')
                    self.open('strong')
                    self.gen_text('Equation')
                    self.close()
                    self.close()
                    self.open('dd')
                    self.open('dl')
                    eqconds_done = 0
                    for asp in a.find_aspects('*'):
                        t = asp.d_tag
                        if t == 'description':
                            self.gen_description_dd(asp)
                        elif not eqconds_done:
                            eqconds_done = 1
                            self.open('dt')
                            cs = a.find_aspects(
                                'precondition', 'postcondition')
                            for cr in cs:
                                if cr.d_tag == 'precondition':
                                    self.open('strong')
                                    self.open('sup')
                                    self.gen_text('pre:')
                                    self.close()
                                    self.close()
                                elif cr.d_tag == 'postcondition':
                                    self.open('strong')
                                    self.open('sup')
                                    self.gen_text('post:')
                                    self.close()
                                    self.close()
                                gen_condition_ref(cr)
                                if cr is not cs[-1]:
                                    self.open('big')
                                    self.gen_text(' == ')
                                    self.close()
                            self.close()
                    self.close()
                    self.close()
                    self.close()
                    self.close()

                else:
                    if not args_described:
                        self.open('dd')
                        self.open('dl')
                        gen_arguments_descriptions(m)
                        args_described = 1
                        self.close()
                        self.close()
                    else:
                        t = last_t
                last_t = t

        self.level += 1
        gen_description(m)
        self.level -= 1

    def gen_getitem_def(self, op):
        self.open('dt')
        self.gen_getitem_ref(op, 1)
        self.close()
        self.gen_mapping_description(op)

    def gen_getitem_ref(self, op, subdescript=0):
        link_name = self.get_link_name(op)

        if subdescript:
            self.open('define', link_name)

        if subdescript:
            self.close()
        else:
            self.open('link_to', link_name)
            self.gen_text('# ')
            self.close()

        self.gen_self(op)
        self.gen_index(op)
        self.gen_returns(op, subdescript)

    def gen_index(self, op):
        self.gen_text('[')
        self.gen_arguments(op.get_arguments())
        self.gen_text(']')

    def gen_link_to_operator(self, link_name):
        self.open('to_html_only')
        self.open('link_to', link_name)
        self.gen_text('# ')
        self.close()
        self.close()

    def gen_mapping_doc(self, m):
        def gen_synopsis(m):
            self.gen_outer_dt('Synopsis')
            self.open('dd')
            self.gen_mapping_ref(m)
            self.close()

        self.gen_header(m)
        self.open('dl')
        self.gen_Name(m)
        gen_synopsis(m)
        self.gen_mapping_description(m)
        self.close()

    def gen_mapping_kind(self, m, subdescript=0, withself=0):
        if withself and self.get_self_node(m):
            self.gen_self(m)
        self.gen_text('(')
        self.gen_arguments(m.get_arguments())
        self.gen_text(')')
        self.gen_returns(m, subdescript)

    def gen_mapping_ref(self, m):
        self.gen_mapping_tag(m)
        self.gen_mapping_kind(m, 0, 1)

    def gen_mapping_tag(self, a, subdescript=0):
        link_name = self.get_link_name(a)
        if not subdescript:
            self.open('link_to', link_name)
        else:
            self.open('define', link_name)

        self.open('strong')
        self.gen_text('callable')
        self.close()
        self.close()
        self.gen_colon()

    def gen_Name(self, m):
        self.gen_outer_dt('Name')
        self.open('dd')
        self.open('h2')
        self.gen_text(m.get_Name())
        self.close()
        self.close()

    def gen_name(self, a):
        self.gen_text(a.get_name())

    def gen_operator_def(self, op):
        self.open('dd')
        self.gen_operator_ref(op, 1)
        self.open('dl')
        self.gen_mapping_description(op)
        self.close()
        self.close()

    def gen_operator_ref(self, op, subdescript=0):
        # self.gen_text('(')

        link_name = self.get_link_name(op)

        if subdescript:
            self.open('define', link_name)

        if subdescript:
            self.close()
        else:
            self.gen_link_to_operator(link_name)

        self.gen_self(op)
        self.gen_text(' ')
        self.open('big')
        self.open('strong')
        self.gen_text(op.src.node.arg.strip())
        self.close()
        self.close()

        for a in op.get_arguments():
            self.gen_text(' ')
            t = a.d_tag
            if t == 'arg':
                self.gen_argref(a)
            else:
                assert 0

        # self.gen_text(')')

        self.gen_returns(op, subdescript)

    def gen_outer_dt(self, text):
        # Synopsis, Description etc

        self.open('dt')
        if not self.level:
            self.open('h2')
        else:
            self.open('strong')
            if self.level == 1:
                self.open('big')
        if callable(text):
            text()
        else:
            self.gen_text(text)
        if self.level == 1:
            self.close('big')
        self.close()
        self.close()

    def gen_posarg_name(self, a):
        self.open('var')
        self.gen_name(a)
        self.close()

    def gen_ref(self, k):
        t = k.d_tag
        getattr(self, 'gen_%s_ref' % t)(k)

    def gen_returns(self, m, subdescript):
        if self.no_ret:
            return
        ars = m.find_aspects('returns')
        if not ars:
            return
        self.open('spc_mapsto')
        self.close()
        rk = m.get_return_kind()
        if subdescript:
            t = rk.d_tag
            if t in ('mapping',):
                self.gen_kind_aspects([rk], 1)
            else:
                self.gen_def(rk)
        else:
            self.gen_ref(rk)

    def gen_reverse_operator_def(self, op):
        self.open('dd')
        self.gen_reverse_operator_ref(op, 1)
        self.open('dl')
        self.gen_mapping_description(op)
        self.close()
        self.close()

    def gen_reverse_operator_ref(self, op, subdescript=0):
        # self.gen_text('(')

        link_name = self.get_link_name(op)

        if subdescript:
            self.open('define', link_name)
            self.close()
        else:
            self.gen_link_to_operator(link_name)

        for a in op.get_arguments():
            t = a.d_tag
            if t == 'arg':
                self.gen_argref(a)
            else:
                assert 0

        self.open('big')
        self.gen_text(' ')

        self.open('strong')
        self.gen_text(op.src.node.arg.strip())
        self.close()

        self.gen_text(' ')
        self.close()

        self.gen_self(op)

        # self.gen_text(')')

        self.gen_returns(op, subdescript)

    def gen_self(self, op):
        s = self.get_self_node(op)
        if not s:
            self.open('em')
            self.gen_text('self')
            self.close()
        else:
            self.out.append(s)

    def gen_self_def(self, k):
        pass

    def gen_self_descriptions(self, li):
        pass

    def gen_self_ref(self, k):
        self.open('h3')
        self.gen_text('For any object ')
        self.open('var')
        self.gen_text(k.src.node.arg.strip())
        self.close()
        self.gen_text(' of kind ')
        self.gen_localname(k.env)
        self.gen_text(':')
        self.close()

    def gen_self_syn(self, li):
        for k in li:
            self.open('dt')
            self.gen_self_ref(k)
            self.close()

    def gen_setitem_def(self, op):
        self.open('dt')
        self.gen_setitem_ref(op, 1)
        self.close()
        self.gen_mapping_description(op)

    def gen_setitem_ref(self, op, subdescript=0):
        link_name = self.get_link_name(op)

        if subdescript:
            self.open('define', link_name)

        if subdescript:
            self.close()
        else:
            self.gen_link_to_operator(link_name)

        self.gen_self(op)
        args = op.get_arguments()
        self.gen_text('[')
        self.gen_arguments(args[:-1])
        self.gen_text(']')

        self.open('strong')
        self.open('big')
        self.gen_text(' = ')
        self.close()
        self.close()

        self.gen_arguments(args[-1:])

    def gen_subkind_of_def(self, k):
        ds = k.find_aspects('description')
        if not ds:
            return
        self.open('dt')
        self.gen_subkind_of_ref(k, 1)
        self.close()
        self.level += 1
        for d in ds:
            self.gen_description_dd(d)

        self.level -= 1

    def gen_subkind_of_descriptions(self, li):
        for a in li:
            self.gen_outer_dt(lambda: self.gen_subkind_of_ref(a, 1))
            for d in a.find_aspects('description'):
                self.gen_description_dd(d)

    def gen_subkind_of_ref(self, k, subdescript=0):
        link_name = self.get_link_name(k)
        if subdescript:
            self.open('define', link_name)
        else:
            self.open('link_to', link_name)
        self.gen_text('Subkind of')
        self.close()
        self.gen_colon()

        comma = 0
        for a in k.args:
            if comma:
                self.gen_text(', ')
            comma = 1
            self.gen_localname(a)

    def gen_subkind_of_syn(self, li):
        for a in li:
            self.gen_outer_dt(lambda: self.gen_subkind_of_ref(a, 0))

    def gen_superkind_of_ref(self, k):
        kas = k.find_kind_aspects()
        if len(kas) == 1:
            self.gen_ref(kas[0])
        else:
            assert 0  # to be tested

    def gen_superkind_ref(self, k):
        self.gen_localname(k)

    def gen_synopsis(self, m):
        ats = m.find_aspects('*')
        ats = self.combine_attrs_of_same_kind(ats)
        tab = self.sortup_aspects(ats, synopsis=1)
        if tab:
            self.gen_outer_dt('Synopsis')
            self.open('dd')
            self.open('dl')
            self.level += 1
            for typ, li in tab:
                try:
                    gen_syn = getattr(self, 'gen_%s_syn' % typ)
                except AttributeError:
                    name = typ.capitalize().replace('_', ' ')
                    if len(li) != 1:
                        name = name+'s'
                    self.gen_outer_dt(name)
                    for a in li:
                        self.open('dd')
                        self.gen_ref(a)
                        self.close()
                else:
                    gen_syn(li)

            self.level -= 1
            self.close()
            self.close()

    def get_self_node(self, a):
        sn = self.use_self
        if sn is None:
            sn = a.get_self_name()
            if sn is not None:
                sn = self.node_of_taci('text', sn)
        return sn

    def sortup_aspects(self, ats, synopsis=0):
        # Get aspects sorted up in the same order for synopsis and main description

        order = ('description', 'subkind_of', 'constructor', 'self', 'method',
                 'operator', 'mapping', 'attribute', 'condition', 'example', )
        tab = {}

        def gen_outer(what):
            assert what in order
            if what not in tab:
                tab[what] = []
            tab[what].append(a)

        for a in ats:
            t = a.d_tag
            if t == 'comment' or t == 'description' and synopsis:
                pass
            elif t in ('attribute', 'attributes', 'either'):
                if a.is_method():
                    gen_outer('method')
                else:
                    gen_outer('attribute')
            elif a.d_type == 'operator' or t in ('getitem', 'delitem', 'setitem'):
                gen_outer('operator')
            else:
                gen_outer(t)

        otab = []
        for typ in order:
            if not typ in tab:
                continue
            li = tab[typ]
            otab.append((typ, li))
        return otab


class _GLUECLAMP_:
    _imports_ = (
        '_parent.FileIO:IO',
        '_parent.Gsml:node_of_gsml',
        '_parent:Html',
        '_parent:Latex',
        '_parent.Main:ReportedError',
        '_parent.SpecNodes:node_of_string',
        '_parent.SpecNodes:node_of_taci',
        '_parent.SpecNodes:node_aliases',
        '_parent:Tester',
        '_parent:XHTML'
    )

    # Map from output mode spelling in output directive to handler name
    # -- Why should we need to map anyway?
    # or should we just say they are case insensitive?
    # If they are case insenitive, we need to map here.
    # I hereby decide they are case insensitive!

    output_handlers = {'html': 'Html', 'xhtml': 'XHTML',
                       'latex': 'Latex', 'tester': 'Tester'}

    # Brackets to use when rendering kind references
    kindbrackets = ('[', ']')
    #kindbrackets = ('{','\n}')
    kindbrackets = (' (', ' )')
    #kindbrackets = '  '

    # Brackets to use when rendering either kinds

    eitherbrackets = '[]'

    # Text to use to render any kind
    anykind = 'anything'

    def document(self, node, env):
        return Document(self, node, env)

    def _test_main_(self):
        class TestSubject:
            def __init__(self, mod, name):
                self.mod = mod
                self.name = name
                self.node_of_string = mod._parent.SpecNodes.node_of_string

            def use(self, args):
                name = self.name
                if name == 'A':
                    return self.node_of_string("""\
.text: hello
""")

                elif name == 'reverse':
                    args.reverse()
                    return self.mod.node_of_taci('string', '', args, 0)
                else:
                    assert 0

        class TestEnv:
            def __init__(self, mod):
                self.mod = mod

            def get_descr_by_name(self, name, node):
                return TestSubject(self.mod, name)

        env = TestEnv(self)
        x = """
.h1: Description of subject
..em
...use: A
.h1: Reversing arguments
.use: reverse
..text: A
..text: B
..text: C
"""
        node = self._parent.SpecNodes.node_of_string(x)
        y = self.document(node, env)
        r = y.get_result()
        # FIXME: Assert equals to something?
        print(r)
        h = self._parent.Html.node2text(r)
        assert h == '''
<h1>Description of subject<em>
 hello
</em></h1>
<h1>Reversing arguments</h1>
 C
 B A'''



##
# Specification entity object
#
# Collects information about a particular system entity or aspect.
#
# The entity is of a general kind
# It is defined by the particular aspects defined for it
# The aspects must be 'compatible', which can be relatively checked.
# The specobject is used for:
# * generating tests
# * generating docs
# * generating help
#   * by creating docstrings
#   * by providing interactive help
#
# For test generation, it will delegate to test implementators.
# For doc generation, it will delegate to doc implementators.
#
# The functionality needed here is therefore limited.
#
#  name
#  aspects
#
# There is one predefine root
# - should we call it Universe?
#
# The name can be full, a dotted name, or the short, last part.
#
# The name is treated as an aspect.
#
# Each aspect definition has a primary kind

# Aspect kinds


from guppy.gsl.Exceptions import *


class SpecObject:
    def __init__(self, mod):
        self.mod = mod

##
# Specification environment
# Collects specifications from several files
# Maps names to specification objects


class SpecEnv:
    def __init__(self, mod):
        self.mod = mod
        self.unknown_nodes = []
        self.files = []

    def visit_default(self, node):
        print('add_unknown', node.tag)
        self.unknown_nodes.append(node)

    def visit_file(self, node):
        print('visit_file')
        file = FileEnv(self, node)
        self.files.append(file)

    def get_predefined_subjects(self, env):
        return (GuppyWorld(env),)


class FileEnv:
    def __init__(self, env, node):
        mod = env.mod
        self.mod = mod
        self.name = self.filename = node.arg
        self.subjects = {}
        for s in env.get_predefined_subjects(self):
            self.subjects[s.name] = s

        file = Subject(self, node, self.name)

        node.children_accept(file)

    def visit_aspects_of(self, node):
        name = node.arg
        subject = self.find_subject(node, name)
        subject.add_aspects(node)

    def def_subject(self, node, name, subject):
        if name in self.subjects:
            self.error_node(node, 'Redefinition of %r.' % name)
            self.error_node(self.subjects[name].node,
                            'Previous definition of %r.' % name)
        else:
            self.subjects[name] = subject

    def error_node(self, node, msg, exception=None):
        index = node.index
        lineno = index + 1
        print('%s:%s:' % (self.filename, lineno))
        print('    %r' % self.get_line(index))
        print('    %s' % msg)
        print()

    def find_subject(self, node, name):
        subject = self.subjects.get(name)
        if subject is None:
            self.error_node(node, 'No such subject: %r.' % name)
        return subject

    def get_line(self, index):
        try:
            with open(self.filename) as f:
                text = list(f.readlines())[index].rstrip()
        except Exception:
            text = None
        return text

    def get_subject(self, name):
        subject = self.subjects.get(name)
        if subject is None:
            subject = self.subjects[name] = Subject(self, name)
        return subject

    def get_aspect_subject(self, env, node):
        name = env.name+'::'+node.tag
        return self.get_subject(name)


class Subject:
    def __init__(self, file, node, name):
        self.file = file
        self.node = node
        self.name = name
        self.aspects = []

    def visit_default(self, node):
        of = node.tag.endswith('_of')
        name = node.arg
        define = name.startswith(':')
        if define:
            if of:

                self.file.error_node(node, "Both 'of' and '::'.")
            name = name[1:].strip()
        if of:
            ofsubject = self.file.find_subject(node, name)
            subject = self.new_subject_of(node, ofsubject)
        else:
            subject = Subject(self.file, node, name)
            if define:
                self.file.def_subject(node, name, subject)
        self.aspects.append(subject)
        node.children_accept(subject)

    def new_subject_of(self, node, of):
        tag = node.tag
        if tag == 'aspects_of':
            return AspectsOf(self.file, node, of)
        else:
            return SubjectOf(self.file, node, of)


class AspectsOf(Subject):
    def __init__(self, file, node, of):
        self.node = node
        self.of = of
        self.aspects = []

    def visit_default(self, node):
        self.of.visit_default(node)


class SubjectOf(Subject):
    def __init__(self, file, node, of):
        self.node = node
        self.of = of
        self.aspects = []


class GuppyWorld(Subject):
    def __init__(self, env):
        self.file = env
        self.name = "Guppy World"
        self.node = None
        self.aspects = []


##
# A node represented with argument splitted in components of the form:
# .tag: arg
# text
# ..child
# ...
# ..child
# ...
#
# @param tag the text of the first line before the colon
# @param arg the text of the first line after the colon (stripped)
# @param text the text after the the first line before the first children
# @param children the child nodes
# @param index line index
# @param src describes the source

class SpecNode(object):
    __slots__ = 'tag', 'arg', 'children', 'index', 'src'

    def __init__(self, tag, arg, children=(), index=0, src=None):
        self.tag = tag
        self.arg = arg
        self.children = tuple(children)
        self.index = index
        self.src = src

    def __repr__(self):
        return '%s(%r,%r,%r)' % (
            self.__class__.__name__, self.tag, self.arg, self.children)

    def __str__(self):
        return '%s(%r,%r,%s)' % (
            self.__class__.__name__, self.tag, self.arg,
            '(%s)' % (','.join([str(c) for c in self.children])))

    def arg_accept(self, visitor, prefix='visit_'):
        if self.arg:
            node = SpecNode('text', self.arg, (), self.index)
            node.accept(visitor, prefix)
        self.children_accept(visitor, prefix)

    def copy(self, tag=None, arg=None, children=None, index=None, src=None):
        if tag is None:
            tag = self.tag
        if arg is None:
            arg = self.arg
        if children is None:
            children = self.children
        if index is None:
            index = self.index
        if src is None:
            src = self.src
        return self.__class__(tag, arg, children, index, src)

    def children_accept(self, visitor, prefix='visit_'):
        for c in self.children:
            c.accept(visitor, prefix)

    def accept(self, visitor, prefix='visit_'):
        m = getattr(visitor, (prefix+self.tag), None)
        if m is None:
            m = getattr(visitor, (prefix+'default'), None)
            if m is None:
                msg = 'accept: unknown: %r, %r  in %r' % (
                    prefix, self.tag, visitor)
                print(msg)
                raise ValueError(msg)
                return
        m(self)

    def error(self, msg, node=None):
        if node is None:
            node = self
        node.src.error(msg, node)

    def get_text(self):
        " Get the total text of all text children, joined with and ended with '\n' "
        text = []
        for c in self.children:
            if c.tag == 'text':
                text.append(c.arg)
                if not c.arg.endswith('\n'):
                    text.append('\n')
        return ''.join(text)

    def get_arg(self):
        arg = self.arg.strip()
        if arg.startswith(':'):
            arg = arg[1:].strip()
        return arg

    def get_arglist(self):
        arg = self.arg
        if arg.startswith(':'):
            arg = arg[1:]
        names = [x.strip() for x in arg.split(',')]
        if names == ['']:
            names = []
        return names

    def get_arg_children(self):
        if self.arg:
            children = [SpecNode('text', self.arg, (), self.index, self.src)]
            children.extend(self.children)
        else:
            children = self.children
        return children

    def get_arg_rest(self, nostrip=0):
        arg = self.arg
        if not nostrip:
            arg = arg.strip()
        return arg, self.children

    def get_arg_norest(self):
        ''' Get the arg as by self.arg,
            but make sure there are no more children.
            '''
        if self.children:
            raise SyntaxError('No children nodes expected in node: %s' % self)
        return self.arg.strip()

    def get_namearg(self):
        ''' Get the argument in the form of a name
            It is the argument stripped.
            And not allowed to contain : or , or new line.
        '''
        name = self.arg.strip()
        if '\n' in name or ':' in name or ',' in name:
            raise SyntaxError('Invalid name: %r' % name)
        return name

    def split_attrs(self, tag=None, attrdict=False):
        if tag is None:
            tag = self.tag
        if attrdict:
            attrs = {}

            def addattr(tag, attr, node):
                if tag in attrs:
                    node.error('Duplicate attribute: %s' % attr)
                else:
                    attrs[tag] = attr
        else:
            attrs = []

            def addattr(tag, attr, node):
                attrs.append((tag, attr))
        children = []
        for ch in self.children:
            if ch.tag == "with":
                for opt in ch.children:
                    if opt.arg:
                        arg = opt.arg
                    else:
                        self.error('Bad attribute, no argument.', opt)
                    if opt.children:
                        self.error(
                            'Expected no children to attribute.', opt.children[0])
                    if opt.arg:
                        addattr(opt.tag, arg, opt)
            elif ch.tag[-1:] == '=':
                addattr(ch.tag[:-1], ch.arg, ch)
            else:
                children.append(ch)
        if len(children) == len(self.children):
            node = self
        else:
            node = self.__class__(
                tag, self.arg, children, self.index, self.src)
        return node, attrs


class Source:
    def __init__(self, name, lines=None, string=None, nostrip=0, debug=0, max_errors=10):
        self.filename = name
        self.lines = lines
        self.string = string
        self.nostrip = nostrip
        self.debug = debug
        self.error_reports = []
        self.max_errors = max_errors
        self.num_warnings = 0
        self.num_errors = 0

    def errmsg_context(self, context):
        linetext = ''
        filename = '<unknown file>'
        if context is not None:
            if hasattr(context, 'index'):
                index = context.index
                src = context.src
            else:
                index = context
                src = self
            if src is not None:
                filename = src.filename
                linetext = src.get_line(index=index)
            print('%s:%s:' % (filename, index+1))
            if linetext:
                print('    %s' % linetext)

    def error(self, message, context=None, exception=None, more=(), harmless=0):
        self.error_reports.append(
            (message, context, exception, more, harmless))
        if harmless:
            self.num_warnings += 1
        else:
            self.num_errors += 1

        self.errmsg_context(context)
        if harmless:
            print('*   %s' % message)
        else:
            print('*** %s' % message)
        print()

        for msg, ctx in more:
            self.errmsg_context(ctx)
            print('    %s' % msg)
            print()

        if self.debug:
            import pdb
            pdb.set_trace()
        else:
            if self.num_errors >= self.max_errors:
                raise TooManyErrors('Too many errors, giving up')
            if exception is not None:
                raise exception

    def get_line(self, index):
        if self.lines is None:
            if self.string is None:
                if self.filename:
                    try:
                        with open(self.filename) as f:
                            self.string = f.read()
                    except Exception:
                        return ''
                else:
                    return ''
            self.lines = self.string.split('\n')
        return self.lines[index]


class _GLUECLAMP_:

    _chgable_ = 'nodemap', 'SpecNode'

    _imports_ = (
        '_parent:DottedTree',
        '_root:re',
        '_root:os',
    )

    node_aliases_defs = (
        ('attr', 'attribute'),
        ('c', 'comment'),
        ('cond', 'condition'),
        ('d', 'description'),
        ('dwh', 'description_with_header'),
        ('eg', 'example'),
        ('fop', 'function_operator'),
        ('iop', 'inplace_operator'),
        ('ka', 'key_arg'),
        ('op', 'operator'),
        ('rop', 'reverse_operator'),
        ('t', 'text'),
    )

    def _get_node_aliases(self):
        return dict(self.node_aliases_defs)

    def _get_reverse_node_aliases(self):
        # Used to make names shorter
        return dict([(v, k) for k, v in self.node_aliases_defs])

    def _get_is_not_ascii(self):
        return self.re.compile(eval(r'u"[\u0080-\uffff]"')).search

    ##
    # @return A tuple of predefined subjects.

    def get_predefined_subjects(self):
        return (GuppyWorld(self),)

    ##
    # Parses a file and makes a tree of nodes
    # @param file name of file containing a dotted tree
    # @return a SpecNode object
    # @more
    # First tag is special.
    # We don't interpret the first line of the file,
    # but uses a special file tag.

    def node_of_file(self, file, nostrip=0):
        src = Source(name=file, nostrip=nostrip)
        dtree = self.DottedTree.parse_file(file, src)
        tag = 'file'
        arg = file
        text = dtree.tag.strip()
        children = self.nodes_of_dforest(dtree.children, src)
        index = dtree.index
        return self.node_of_tatci(tag, arg, text, children, index, src)

    def node_of_string(self, string, name='<string>', nostrip=0):
        dtree = self.DottedTree.parse_string(string)
        tag = 'string'
        arg = name
        src = Source(name=name, string=string, nostrip=nostrip)
        text = dtree.tag.strip()
        index = dtree.index
        children = self.nodes_of_dforest(dtree.children, src)
        return self.node_of_tatci(tag, arg, text, children, index, src)

    def node_of_dtree(self, dtree, src):
        tag = dtree.tag
        textpos = tag.find('\n')
        if textpos == -1:
            textpos = len(tag)

        equpos = tag.find('=', 0, textpos)
        colonpos = tag.find(':', 0, textpos)
        if equpos != -1 and (colonpos == -1 or equpos < colonpos):
            tag, arg = (tag[:equpos].strip()+'=',
                        tag[equpos+1:].strip())
        else:
            if colonpos == -1:
                if not ' ' in tag[:textpos] or textpos >= len(tag.rstrip()):
                    colonpos = textpos
                else:
                    raise SyntaxError(
                        'No colon in spaced tag in node %s' % dtree)
            tag, arg = (tag[:colonpos].strip(),
                        tag[colonpos+1:]
                        )
        if tag in self.node_aliases:
            tag = self.node_aliases[tag]
        tag = tag.replace(' ', '_')
        if tag != 'text' and not src.nostrip:
            arg = arg.strip()

        children = self.nodes_of_dforest(dtree.children, src)
        return self.node_of_taci(tag, arg, children, dtree.index, src)

    def nodes_of_dforest(self, dforest, src):
        onodes = [self.node_of_dtree(c, src) for c in dforest]
        nodes = []
        for node in onodes:
            if node.tag != 'include':
                nodes.append(node)
                continue
            filename = node.arg.strip()
            filename = self.os.path.join(self.os.path.dirname(src.filename),
                                         filename)
            node = self.node_of_file(filename, nostrip=src.nostrip)
            nodes.extend(node.children)
        return tuple(nodes)

    def _get_node_of_taci(self):
        return SpecNode

    def node_of_tatci(self, tag, arg, text, children=(), index=0, src=None):
        if text:
            if tag == 'text':
                if arg:
                    arg = arg + '\n' + text
                else:
                    arg = text
            else:
                children = (self.node_of_taci(
                    'text', text, (), index, src),) + children
        return self.node_of_taci(tag, arg, children, index, src)

    def node_of_text(self, text):
        # Returns a node that is either
        # - a 'text' node, if text was all ascii
        # - a 'char' node, if text was a single non-ascii
        # - a 'block' with children being a sequence of char and text nodes,
        #       if text contained ascii and non-ascii characters
        nodes = self.nodes_of_text(text)
        if len(nodes) == 1:
            return nodes[0]
        else:
            return self.node_of_taci('block', '', nodes)

    def nodes_of_text(self, text):
        # Returns a sequence of nodes, encoding text.
        nodes = []
        if self.is_not_ascii(text):
            chars = []
            for char in text:
                no = ord(char)
                if no < 128:
                    chars.append(char)
                else:
                    if chars:
                        nodes.append(self.node_of_taci('text', ''.join(chars)))
                        chars = []
                    nodes.append(self.node_of_taci('char', str(no)))
            if chars:
                nodes.append(self.node_of_taci('text', ''.join(chars)))
        else:
            nodes.append(self.node_of_taci('text', text))
        return nodes

    def main(self):
        root = self._root

        specdir = root.os.path.join(
            root.os.path.dirname(root.os.path.realpath(__file__)),
            '../../specs')
        main_dt_name = root.os.path.join(specdir, "docexample.gsl")
        if not root.os.path.exists(main_dt_name):
            print('%s does not exist, skipping test' % main_dt_name)
            return

        env = self.SpecEnv(self)

        node = self.node_of_file(main_dt_name)
        node.accept(env)

    def print_doc(self, dt):
        self.print_doc()

    def unparse_head(self, level, tag, arg, text):
        head = tag
        if arg:
            head = head + ': ' + arg
        if text:
            head = head + '\n' + text
        tag = self.DottedTree.unparse_tag(level, head)
        return tag


def test_main():
    from guppy import Root
    root = Root()

    root.guppy.gsl.SpecNodes.main()



from guppy.etc.Descriptor import property_nondata
from guppy.gsl.Exceptions import *


class SpecEnv:
    def __init__(self, mod):
        self.mod = mod
        self.imported_packages = {}
        self.importing_packages = {}
        self.error_reports = []
        self.num_errors = 0
        self.num_warnings = 0

    def errmsg_context(self, context):
        linetext = ''
        filename = '<unknown file>'
        if context is not None:
            node = context      # Assume it's a node - that's all we use for now
            lineno = node.index + 1
            src = node.src
            if src is not None:
                filename = src.filename
                linetext = src.get_line(index=context.index)
            print('%s:%s:' % (filename, lineno))
            if linetext:
                print('    %r' % linetext)

    def error(self, message, context=None, exception=ReportedError, more=(), harmless=0):
        self.error_reports.append(
            (message, context, exception, more, harmless))
        if harmless:
            self.num_warnings += 1
        else:
            self.num_errors += 1

        self.errmsg_context(context)
        if harmless:
            print('*   %s' % message)
        else:
            print('*** %s' % message)
        print()

        for msg, ctx in more:
            self.errmsg_context(ctx)
            print('    %s' % msg)
            print()

        if self.debug:
            import pdb
            pdb.set_trace()
        else:
            if self.num_errors >= self.max_errors:
                raise TooManyErrors('Too many errors, giving up')
            if exception is not None:
                raise exception

    def get_filers(self, documents):
        filers = []
        for d in documents:
            filers.extend(d.get_filers(self.output_dir))
        return filers

    def import_package(self, name, context):
        pac = self.imported_packages.get(name)
        if pac is None:
            if name in self.importing_packages:
                self.error('Invalid mutual import involving packages %r' % (
                    list(self.importing_packages.keys()),), context)
            self.importing_packages[name] = 1
            filename = name.replace('.', self.mod.IO.path.sep)+'.gsl'
            ip = self.package_of_filename(filename, name)
            pac = self.mkPackage(ip)
            self.imported_packages[name] = pac
            del self.importing_packages[name]
        return pac

    def link_documents(self, documents):
        defines = {}

        links = {}

        def walk(node):
            t = node.tag
            if t == 'link_to':
                name = node.arg.strip()
                links.setdefault(name, []).append((d, node))
            elif t == 'define':
                name = node.arg.strip()
                defines.setdefault(name, []).append((d, node))
            elif t == 'to_tester_only':
                return
            for ch in node.children:
                walk(ch)

        for d in documents:
            node = d.get_result()
            walk(node)

        for name, ds in list(defines.items()):
            if len(ds) > 1:
                print('Duplicate definition of name %r, defined in:' % name)
                for (d, node) in ds:
                    print('    %s line %s' % (d.get_doc_name(), node.index+1))
                print('Will use the first one.')

        nodefs = []

        for name, ds in list(links.items()):
            if name not in defines:
                used = {}
                for (d, node) in ds:
                    used[d.get_doc_name()] = 1
                    node.tag = 'link_to_unresolved'
                used = list(used.keys())
                used.sort()
                used = ', '.join(used)
                nodefs.append('%s used in %s' % (name, used))
            else:
                defd, defnode = defines[name][0]
                for (d, node) in ds:
                    if d is defd:
                        node.tag = 'link_to_local'
                    else:
                        node.tag = 'link_to_extern'
                        node.children = (defd.doc_name_node,)+node.children
        if nodefs:
            nodefs.sort()
            print('Unresolved links:')
            for nd in nodefs:
                print('  ', nd)

    def mkPackage(self, sub):
        pac = PackageDescription(self, sub, sub)
        pac.output_dir = self.output_dir
        pac.resolve_all()
        return pac

    def package_of_filename(self, filename, packname=None, nostrip=1, input_string=None):
        mod = self.mod
        if packname is None:
            if filename.endswith('.gsl'):
                packname = filename[:-4]
            else:
                packname = filename
            packname = packname.replace(mod.IO.path.sep, '.')
        if self.input_dir:
            filename = mod.IO.path.join(self.input_dir, filename)
        else:
            filename = mod.IO.path.abspath(filename)
        if input_string is not None:
            data = input_string
        else:
            data = mod.IO.read_file(filename)
        md5 = mod.md5()
        md5.update(b'.filename: %s\n' % filename.encode('utf-8'))
        md5.update(b'.packname: : %s\n' % packname.encode('utf-8'))
        md5.update(data.encode('utf-8'))
        digest = md5.digest()
        if digest in mod.package_cache:
            return mod.package_cache[digest]

        node = mod.SpecNodes.node_of_string(data, filename, nostrip=nostrip)
        numerr = self.num_errors
        print('Making package subject %r' % packname)
        package = PackageSubject(mod, self, node, packname, filename)
        if numerr == self.num_errors:
            mod.package_cache[digest] = package
        return package

    def process_main(self, filename, input_dir=None, output_dir=None, debug=False, max_errors=None,
                     process_despite_errors=False, raise_at_errors=False,
                     input_string=None
                     ):
        if input_dir is None:
            input_dir = self.mod.input_dir
        self.input_dir = input_dir
        if output_dir is None:
            output_dir = '/tmp'
        self.output_dir = output_dir
        self.debug = debug
        if max_errors is None:
            max_errors = self.mod.max_errors
        self.max_errors = max_errors

        try:
            pac = self.mkPackage(self.package_of_filename(
                filename, input_string=input_string))
            documents = pac.get_documents()
            if not documents:
                self.error('No documents specified.',
                           exception=None, harmless=1)
            if not self.num_errors or process_despite_errors:
                print('Linking')
                self.link_documents(documents)
            if not self.num_errors or process_despite_errors:
                filers = self.get_filers(documents)
        except TooManyErrors:
            giving_up = ' giving up --'
        else:
            giving_up = ''
        if not self.num_errors:
            for filer in filers:
                f = self.mod.Filer.filer(filer)
                print('Writing: ', ', '.join(list(f.writefile_names)))
                f.write()
        if self.num_warnings:
            print('*   %d warning%s reported.' % (
                self.num_warnings, 's'[:self.num_warnings > 1]))
        if self.num_errors:
            print('*** %d error%s reported --%s no files written.' % (
                self.num_errors, 's'[:self.num_errors > 1], giving_up))
            if raise_at_errors:
                raise HadReportedError('Some error has been reported.')


class UntypedDescription:
    def __init__(self, env, tgt, src):
        self.env = env
        self.pac = env.pac
        self.mod = env.mod
        self.tgt = tgt
        self.src = src

    def combine_with_subject(self, subject):
        self.combined_subjects.append(subject)

    def resolve_all(self):
        self.resolve_primary()
        self.resolve_lookuped()

    def resolve_primary(self):
        self.resolve_type()
        self.resolve_tgt()

    def resolve_type(self):
        dc = self.tgt.description_class
        if not hasattr(dc, 'd_tag'):
            self.d_tag = self.tgt.tag
        self.tgtfullname = self.mod.tgt_prefix+self.tgt.fullname
        self.tgtnode = self.tgt.node
        self.tgtlastname = self.tgt.lastname
        self.srcnode = self.tgt.node
        self.srcfullname = self.src.fullname
        self.srclastname = self.tgt.lastname

        self.__class__ = dc


class Description:
    d_max_occur = None  # Max occurence as an aspect if a number
    d_sub = ()  # Tags of allowed sub-aspects
    d_type = 'other'
    d_is_def = 0
    is_lookuped = False
    is_synthetic = False  # Set if it was made not to correspond with a user node
    the_less_specific_descr = None
    args = ()

    def aspects_extend(self, as_):
        for asp in as_:
            try:
                k = asp.src.definame
                # k = asp.tgt.definame # Humm
                if k:
                    w = self.localview.get(k)
                    if w:
                        if w is asp:
                            # May happen eg as in test16, for a product
                            # But it is somewhat mystical.
                            continue
                        self.error('Duplicate aspect %r (may be correct in future).' % (k,),
                                   w.src.node,
                                   DuplicateError)
                    self.localview[k] = asp
                bn = self.aspects_by_tag.setdefault(asp.d_tag, [])
                oc = asp.d_max_occur
                if oc is not None:
                    if len(bn) + 1 > oc:
                        self.error('More than %d %r aspects.' % (
                            oc, asp.d_tag), asp.src.node)
                bn.append(asp)
                self.aspects.append(asp)
            except ReportedError:
                pass

    def aspects_extend_by_subjects(self, subjects):
        for v in subjects:
            try:
                asp = UntypedDescription(self, v, v)
                asp.resolve_primary()
                self.aspects_extend((asp,))
            except ReportedError:
                pass

    def deftgt(self, forme=None):
        if forme is None:
            forme = self

        try:
            tgtview = self.tgtview
        except AttributeError:
            self.env.deftgt(forme)
        else:
            if forme.tgtfullname in tgtview:
                self.error('Duplicate definition of %r' %
                           forme.tgtfullname, forme.src.node)
            tgtview[forme.tgtfullname] = forme

    def error(self, msg, node=None, exception=ReportedError, **kwds):
        return self.pac.env.error(msg, node, exception, **kwds)

    def resolve_lookuped(self):
        if not self.is_lookuped:
            self.is_lookuped = 1
            self.resolve_aspects()

    def resolve_tgt(self):
        self.deftgt()

    def find_aspects(self, tag='*', *tags):
        al = []
        tag = tag.replace(' ', '_')
        if tag in ('*', 'arg'):
            for a in self.args:
                a.resolve_lookuped()
                al.append(a)
            if tag == '*':
                for a in self.aspects:
                    a.resolve_lookuped()
                    al.append(a)
                return al
        tags = (tag,) + tags
        for a in self.aspects:
            if a.d_tag in tags:
                a.resolve_lookuped()
                al.append(a)
        return al

    def find_arg_aspects(self):
        al = []
        for a in self.args:
            a.resolve_lookuped()
            al.append(a)
        for a in self.aspects:
            if a.d_tag in ('arg', 'seq', 'repeat', 'alt', 'args', 'optionals', 'key_arg',
                           'draw', 'no_arg'):
                a.resolve_lookuped()
                al.append(a)
        return al

    def find_kind_aspects(self):
        kas = []
        for asp in self.find_aspects('*'):
            if asp.d_tag in ('attribute', 'mapping', 'kind', 'either', 'kind_of', 'superkind',
                             'superkind_of'):
                kas.append(asp)
            else:
                pass
        return kas

    def merge_policy(self, descrs):
        return descrs

    def get_descr_for_aspect(self, aspect):
        if not self.aspects and self.the_less_specific_descr is not None:
            return self.the_less_specific_descr.get_descr_for_aspect(aspect)
        return self

    def get_atom_beams(self):
        aspects = self.find_aspects('*')
        aks = []
        for asp in aspects:
            if asp.d_tag in ('attribute', 'mapping', 'either',  'operator',
                             'inplace_operator', 'reverse_operator', 'function_operator',
                             'delitem', 'getitem', 'setitem',
                             ):
                aks.append(beam(self, asp))
            elif asp.d_tag in ('kind', 'kind_of', 'subkind_of') and asp is not self:
                a = beam(self, asp)
                for b in asp.get_atom_beams():
                    aks.append(a + b)
        return aks

    def get_aspects_kind(self, aspects=None):
        if aspects is None:
            aspects = self.find_aspects('*')
        aks = []
        for asp in aspects:
            if asp.d_tag in ('attribute', 'mapping', 'either',
                             'operator', 'inplace_operator', 'reverse_operator', 'function_operator',
                             'delitem', 'getitem', 'setitem',
                             ):
                aks.append(asp)
            elif asp.d_tag in ('kind', 'kind_of', 'subkind_of') and asp is not self:
                aks.extend(asp.get_atom_kinds())
        return aks

    def get_atom_kinds(self):
        return self.get_aspects_kind([self] + self.find_aspects('*'))

    def get_examples(self, get_all=False):
        examples = []
        exs = self.find_aspects('example')
        for ex in exs:
            examples.extend(ex.get_examples())
        return examples

    def get_re(self, opt):
        if opt.get('get_examples'):
            exres = [self.mod.RE.Single(x) for x in self.get_examples()]
            if not exres:
                self.error('Test coverage error: no examples specified.',
                           self.tgt.node,
                           CoverageError)
            return self.mod.RE.Union(*exres)
        else:
            return self.mod.RE.Single(self)

    def get_most_specific_descrs(self, descrs):
        nds = []
        for d in descrs:
            nds = [x for x in nds if not d.is_more_specific_than(x)]
            for x in nds:
                if x is d:
                    break
                if x.is_more_specific_than(d):
                    break
            else:
                nds.append(d)
        return nds

    def get_package(self):
        return self.pac

    def is_more_specific_than(self, d):
        r = self.the_less_specific_descr
        return r is d or (r is not None and r.is_more_specific_than(d))

    def get_self_name(self):
        def find(e):
            sa = e.find_aspects('self')
            if sa:
                # length = 1, has been checked
                assert len(sa) == 1
                return sa[0].src.node.arg.strip()
            if e.d_tag != 'package':
                return find(e.env)
            return None
        return find(self)

    def gen_description_doc(self, out):
        ds = self.find_aspects('description')
        if not ds:
            out.gen_text('<NO DESCRIPTION OF %r>' % self.tgtfullname)
        else:
            for d in ds:
                d.gen_doc(out)

    def get_id_name(self):
        return self.tgtfullname

    def get_link_name(self):
        return self.tgtfullname

    def get_local_name(self):
        return self.srclastname

    def get_test_name(self):
        return self.tgtfullname

    def get_name(self):
        return self.tgtfullname

    def get_Name(self):
        # To be used in Name of doc.
        n = self.find_aspects('name')
        if not n:
            name = self.tgtlastname
        else:
            name = n.tgt.node.arg.strip()
        return name

    def get_descr_by_subject(self, subject):
        return self.pac.get_descr_by_subject(subject)

    def init_localview(self, only_vars=0):
        self.localview = {}
        self.aspects = []
        self.aspects_by_tag = {}

        if not only_vars:
            self.aspects_extend_by_subjects(self.tgt.aspects)

    def resolve_aspects(self):
        self.init_localview()
        if self.src.args:
            self.args = [self.env.get_descr_by_subject(
                arg) for arg in self.src.args]
        self.resolve_special()

    def resolve_special(self):
        # To be overridden with special checks etc.
        pass

    def get_the_one_argument(self):
        arg = self.src.node.arg.strip()
        if self.aspects:
            'No children expected for %r' % self.node.tag
        return arg

    def make_and_test_kind(self, kinds):
        ks = []

        def flatten(k):
            if k.d_tag == 'kind':
                for k1 in k.find_kind_aspects():
                    flatten(k1)
            else:
                ks.append(k)

        if (len(kinds) == 1 and kinds[0].d_tag == 'kind'):
            return kinds[0]
        for k in kinds:
            flatten(k)
        kinds = ks

        k = Kind()
        k.d_tag = 'kind'
        k.aspects = kinds
        k.tgtfullname = '(%s)' % ('&'.join([x.tgtfullname for x in kinds]))
        k.is_lookuped = 1
        return k

    def make_and_kind(self, kinds):
        if (len(kinds) == 1 and kinds[0].d_tag in('kind', 'kind_of')):
            return kinds[0]

        k = Kind()
        k.d_tag = 'kind'
        k.aspects = kinds
        k.tgtfullname = '(%s)' % ('&'.join([x.tgtfullname for x in kinds]))
        k.is_lookuped = True
        k.is_synthetic = True
        return k

    def make_or_kind(self, kinds):
        if len(kinds) == 1:
            return kinds[0]
        else:
            k = Superkind()
            k.d_tag = 'kind'
            k.aspects = kinds
            k.tgtfullname = '(%s)' % ('|'.join([x.tgtfullname for x in kinds]))
            k.is_lookuped = True
            k.is_synthetic = True
            return k


class Definition(Description):
    d_is_def = 1
    d_type = 'definition'

    def export_aspects(self, src):
        src.__class__ = self.__class__
        if src.d_tag == 'import':
            src.d_tag = self.d_tag
        else:
            if src.d_tag != self.d_tag:
                # Can't think of how this would happen -
                # so not yet converted to .error()
                raise ImportError('Different description tag')
        src.aspects_extend(self.aspects)


class DescriptionDescription(Description):
    d_sub = ('text', )
    d_tag = 'description'

    def gen_doc(self, out):
        self.srcnode.arg_accept(out)


class Default(DescriptionDescription):
    def gen_doc(self, out):
        arglines = self.srcnode.arg.strip().split('\n')
        default = arglines[0]
        rest = '\n'.join(arglines[1:])
        out.open('dl')
        out.open('dt')
        out.open('strong')
        out.gen_text('Default: ')
        out.close()
        out.gen_text(default)
        out.close()
        out.open('dd')
        out.gen_text(rest)
        self.srcnode.children_accept(out)
        out.close()
        out.close('dl')


class DescriptionWithHeader(DescriptionDescription):
    def gen_doc(self, out):
        arglines = self.srcnode.arg.strip().split('\n')
        header = arglines[0]
        rest = '\n'.join(arglines[1:])

        out.open('dl')
        out.gen_outer_dt(header)
        out.open('dd')
        out.gen_text(rest)
        self.srcnode.children_accept(out)
        out.close()
        out.close()


class Comment(DescriptionDescription):
    d_tag = 'comment'
    pass


class Either(Description):
    d_type = 'with_args'

    def get_atom_beams(self):
        return [beam(self)]

    def get_atom_kinds(self):
        return [self]

    def get_alt_kinds(self):
        return self.find_kind_aspects()


class Import(Definition):
    d_sub = ('from', 'resolve_by', 'using',
             'attribute', 'condition', 'description', 'comment', 'constructor',
             'mapping', 'method',
             'operator', 'inplace_operator', 'reverse_operator', 'function_operator',
             'delitem', 'getitem', 'setitem',
             'self',
             'subkind_of',
             )

    def resolve_tgt(self):
        self.is_lookuped = 1
        using_name, using_node = self.src.imp_using_map.get(
            self.src.definame, (self.src.definame, self.src.node))
        import_node = self.src.node
        ds = [self.pac.import_package(from_name, from_node).
              get_descr_by_name(using_name, using_node)
              for (from_name, from_node) in self.src.imp_froms]

        if len(ds) == 1:
            d = ds[0]
        else:
            d = Product(self, ds, ProductSubject([x.src for x in ds]),
                        self.src.imp_resolve_mode)

        self.tgt = d.tgt
        self.tgtfullname = self.mod.tgt_prefix+self.tgt.fullname
        self.the_less_specific_descr = d

        self.init_localview(only_vars=1)
        d.export_aspects(self)
        self.aspects_extend_by_subjects(self.src.aspects)
        self.deftgt()

    def resolve_aspects(self):
        pass


class Product(Description):
    def __init__(self, env, ds, src, mode):
        self.env = env
        self.mod = env.mod
        self.src = src
        self.mode = mode
        self.pac = env.pac

        tgt = ds[0].tgt
        for d in ds[1:]:
            if d.tgt is not tgt:
                self.error('Import error when importing from multiple packages:\n' +
                           '  Can not make a product of %r (tgt = %r) with %r (tgt = %r)\n' % (
                               d.src.fullname, d.tgt.fullname, ds[0].src.fullname, ds[0].tgt.fullname) +
                           '  because of different targets.',
                           d.src.node)

        self.tgt = tgt
        self.ds = ds

    def export_aspects(self, src):
        for d in self.ds:
            d.export_aspects(src)

    def is_more_specific_than(self, d):
        for x in self.ds:
            if x is d or x.is_more_specific_than(d):
                return True
        return False


class PackageDescription(UntypedDescription):
    def __init__(self, env, tgt, src):
        self.env = env
        self.pac = self
        self.mod = env.mod
        self.tgt = tgt
        self.src = src


class ErrorDescription:
    d_tag = 'error'

    def __init__(self, env):
        self.env = env

    def get_id_name(self):
        return '<error>.<error>'


class Package(Description):
    d_sub = ('and', 'comment', 'condition', 'document', 'import', 'kind', 'macro',
             'superkind',
             )

    def get_tgtdicts(self):
        seen = {id(self.tgtview): 1}
        tgtdicts = [self.tgtview]
        for p in list(self.imported_packages.values()):
            sds = p.get_tgtdicts()
            for sd in sds:
                if id(sd) not in seen:
                    seen[id(sd)] = 1
                    tgtdicts.append(sd)
        return tgtdicts

    def get_descr_by_name(self, name, context=None):
        if name.startswith(self.mod.tgt_prefix):
            return self.get_descr_by_tgt_name(name, context)

        e = self
        parts = name.split('.')
        for part in parts:
            try:
                e = e.localview[part]
            except KeyError:
                assert context
                self.env.error(
                    'Undefined: %r in %r.' % (part, e.get_id_name()), context,
                    exception=UndefinedError)
            e.resolve_lookuped()
        return e

    def get_descr_by_subject(self, subject):
        name = subject.fullname
        if name.startswith(self.srcfullname+'.'):
            name = name[len(self.srcfullname)+1:].strip()
        else:
            self.error('Undefined: %r' % name, subject.node)
        return self.get_descr_by_name(name, subject.node)

    def get_descr_by_tgt_name(self, name, context=None):
        tgtdicts = self.get_tgtdicts()
        descrs = []
        for tgtdict in tgtdicts:
            if name in tgtdict:
                d = tgtdict[name]
                d.resolve_lookuped()
                d = d.get_descr_for_aspect('*')
                descrs.append(d)
        if not descrs:
            self.error('No definition of tgt %r' %
                       name, context, UndefinedError)
        descrs = self.get_most_specific_descrs(descrs)
        if len(descrs) > 1:
            descrs = self.merge_policy(descrs)
            if len(descrs) > 1:
                self.error('Conflicting descriptions of %r:%r' % (
                    name, [d.src.fullname for d in descrs]),
                    context,
                    DuplicateError)

        return descrs[0]

    def get_filename(self):
        return self.src.filename

    def get_package(self):
        return self

    def resolve_tgt(self):
        self.tgtview = {}

    def resolve_aspects(self):
        self.imported_packages = {}
        self.init_localview()

    def import_package(self, name, context):
        pac = self.imported_packages.get(name)
        if pac is None:
            pac = self.env.import_package(name, context)
            self.imported_packages[name] = pac
        return pac

    def get_documents(self):
        documents = []
        for doc in self.src.documents:
            node = doc.node
            doc = self.mod.Document.document(node, self)
            documents.append(doc)

        return documents


class Attribute(Definition):
    d_sub = ('attribute', 'comment', 'description', 'description_with_header',
             'either', 'kind_of', 'mapping', 'method', 'self')

    def export_aspects(self, src):
        src.__class__ = self.__class__
        src.aspects_extend(self.aspects)

    def get_attr_name(self):
        return self.tgtlastname

    def get_name(self):
        return self.tgtlastname

    def get_kind(self):
        kas = self.find_kind_aspects()
        return self.make_and_kind(kas)

    def get_kind_name(self):
        k = self.get_kind()
        if k.d_tag == 'kind_of':
            kas = k.find_kind_aspects()
            if len(kas) == 1:
                k = kas[0]
            else:
                raise ValueError("Don't know how to name this kind, %r" % self)
        return k.tgtfullname

    def get_link_name(self):
        # xxx needs smoother logic
        s = '%s.%s' % (self.get_descr_by_subject(
            self.tgt.parent).get_link_name(), self.tgt.lastname)
        return s

    def get_test_kind(self):
        kas = self.find_kind_aspects()
        return self.make_and_test_kind(kas)

    def is_method(self):
        return (self.find_aspects('mapping') and
                not self.find_aspects('kind_of'))

    def get_op_name(self):
        return self.get_attr_name()


class KindOf(Description):
    d_type = 'with_args'
    d_sub = ()


class SubkindOf(Description):
    d_type = 'with_args'
    d_sub = ('description',)


class Kind(Definition):
    d_sub = ('attribute', 'condition', 'description', 'comment', 'constructor',
             'example',
             'mapping', 'method',
             'operator', 'inplace_operator', 'reverse_operator', 'function_operator',
             'self',
             'subkind_of',
             'delitem', 'getitem', 'setitem',
             )

    def get_attributes(self):
        return self.find_aspects('attribute')

    def get_mappings(self):
        return self.find_aspects('mapping')


class Superkind(Definition):
    d_sub = ('comment', 'description', 'example', 'superkind_of')

    def get_local_name(self):
        return self.srclastname


class SuperkindOf(Description):
    d_type = 'with_args'

    def get_examples(self, enough=1):
        examples = Description.get_examples(self, enough)
        if len(examples) < enough:
            for ka in self.find_kind_aspects():
                if ka is self:
                    continue
                examples.extend(ka.get_examples(enough-len(examples)))
                if len(examples) >= enough:
                    break
        return examples


class Example(Description):
    d_sub = ('comment', 'description', 'in_context')
    partab = {"'''": "'''",
              '"""': '"""',
              '(': ')',
              '[': ']',
              '{': '}'
              }

    def get_ex_text(self):
        return self.src.ex_text

    def get_examples(self, get_all=False):
        return [self]

    def get_ctx_text(self):
        asp = self.find_aspects('in_context')
        if not asp:
            return ''
        # It is of length 1, has been checked.
        return asp[0].tgt.node.arg.strip()

    def get_use_text(self, x):
        return x


class InContext(Description):
    d_max_occur = 1


class Defines(Description):
    d_type = 'with_args'

    def get_defined_tgt_names(self):
        return [x.tgtfullname for x in self.find_aspects('arg')]


class Macro(Definition):
    def export_aspects(self, src):
        src.__class__ = self.__class__
        src.tgtnode = self.tgtnode

    def use(self, options):
        return self.mod.SpecNodes.node_of_taci(
            'block', '', self.tgtnode.children, self.tgtnode.index)


class Self(Description):
    d_max_occur = 1


class Mapping(Description):
    d_type = 'other'
    d_sub = ('alt', 'arg', 'args', 'comment', 'description', 'description_with_header',
             'equation',
             'draw',
             'key_arg',
             'optionals',
             'precondition', 'postcondition',
             'repeat', 'returns',
             'self',
             'seq',
             )

    def chk_num_args(self, min, max):
        re = self.get_args_re({})
        xs = re.sequni()
        for x in xs:
            try:
                if min is not None and min == max and len(x) != min:
                    self.error(
                        '%s requires %d argument%s specified, got %d.' % (
                            self.d_tag, min, 's'[min == 1:], len(x)),
                        self.src.node)

                elif min is not None and len(x) < min:
                    self.error(
                        '%s requires at least %d argument%s specified, got %d.' % (
                            self.d_tag, min, 's'[min == 1:], len(x)),
                        self.src.node)

                elif max is not None and len(x) > min:
                    self.error(
                        '%s can take at most %d argument%s specified, got %d.' % (
                            self.d_tag, max, 's'[max == 1:], len(x)),
                        self.src.node)
            except ReportedError:
                pass

    def get_arg_kinds(self):
        ak = []
        for a in self.find_aspects('args'):
            ak.extend(list(a.args))
        return ak

    def get_args_examples(self, mapname, top_kind):
        # Get arguments example, esp. for test purposes

        try:
            opt = {'get_examples': True}

            re = self.get_args_re(opt)

            coverage = 1
            try:
                xs = re.sequni()
            except self.mod.RE.InfiniteError:
                print('Infinitely long args example for %s' % self.srcfullname)
                print(
                    'Limiting by expanding each Cleene closure 0 up to %d times.' % coverage)
                re = re.limited(coverage)
                xs = re.sequni()
            examples = [ArgsExample(self, tuple(
                x), mapname, top_kind) for x in xs]
        except CoverageError:
            return []
        else:
            return examples

    def get_args_for_args(self, args, match):
        arglist = []
        for a in self.find_arg_aspects():
            t = a.d_tag
            if t == 'arg':
                name = a.get_name()
                if name in match:
                    v = args.get_arg_value(match[name])
                else:
                    ex = a.get_examples()
                    if not ex:
                        # I have been able to cause this to happen in test67.
                        self.error(
                            'Test coverage error: Can not create precondition for %r\n -- no examples specified for the argument above.' % args.mapping.tgtfullname,
                            a.src.node
                        )
                    v = ex[0]
                arglist.append(v)
            else:
                assert 0
                # raise ConditionError, 'Can not match this precondition'

        return ArgsExample(self, tuple(arglist), args.mapname, args.top_kind)

    def get_args_re(self, opt):
        re = self.mod.RE.Epsilon
        for a in self.find_arg_aspects():
            re += a.get_re(opt)
        return re

    def get_arguments(self):
        # Get the arguments subjects, for doc description purposes
        return self.find_arg_aspects()

    def get_return_kind(self):
        return self.make_and_kind([x.get_kind() for x in self.find_aspects('returns')])

    def get_return_test_kind(self):
        return self.make_and_test_kind([x.get_test_kind() for x in self.find_aspects('returns')])


class ArgsExample:
    def __init__(self, mapping, egs, mapname, top_kind):
        self.mapping = mapping
        self.egs = egs
        self.mapname = mapname
        self.top_kind = top_kind
        self.negs = [mapname(x) for x in egs]

    def __str__(self):
        return ', '.join(self.negs)

    def get_arg_value(self, name):
        i = 0
        for a in self.mapping.find_arg_aspects():
            t = a.d_tag
            if t == 'arg':
                if a.get_name() == name:
                    return self.egs[i]
            else:
                raise ConditionError('No argument matches: %r' % name)
            i += 1

    def get_preconditions(self):
        return self.mapping.find_aspects('precondition')

    def get_postconditions(self):
        return self.mapping.find_aspects('postcondition')

    def get_setups_for_preconditions(self):
        pres = self.get_preconditions()
        if not pres:
            return []
        kind = self.top_kind

        map = self.mapping

        pres = map.find_aspects('precondition')
        if pres:
            for a in kind.find_aspects('attribute'):
                for m in a.find_aspects('mapping'):
                    mpre = m.find_aspects('precondition')
                    if mpre:
                        continue
                    match = self.match_to(m.find_aspects('postcondition'))
                    if match is not None:
                        # found one
                        args = m.get_args_for_args(self, match)
                        return [SetUp(a.get_attr_name(), args)]
                        break
                else:
                    continue
                break
            else:
                # Caller will do error reporting
                return None
        return []

    def match_to_kind(self, kind):
        pass

    def match_to(self, posts):
        match = {}
        for pre in self.get_preconditions():
            for pos in posts:
                if pos.cond_id == pre.cond_id:
                    if len(pos.arg_names) != len(pre.arg_names):
                        continue
                    upd = {}
                    for a, b in zip(pos.arg_names, pre.arg_names):
                        if a in match:
                            break
                        upd[a] = b
                    else:
                        match.update(upd)
                        break
            else:
                return None
        assert ',' not in match
        return match


class SetUp:
    def __init__(self, name, args):
        self.name = name
        self.args = args

    def get_name(self):
        return self.name

    def get_args(self):
        return self.args


class Operator(Mapping):
    d_is_def = 1
    d_type = 'operator'
    d_sub = ('arg', 'comment', 'description', 'description_with_header',
             'equation',
             'postcondition', 'precondition',
             'self', 'returns', )

    def get_op_name(self):
        return self.src.node.arg.strip()

    def resolve_special(self):
        self.chk_num_args(1, 1)


class ReverseOperator(Operator):
    pass


class FunctionOperator(Operator):
    def resolve_special(self):
        self.chk_num_args(0, 0)


class InplaceOperator(Operator):
    pass


class SetItem(Mapping):
    d_type = 'other'
    d_sub = ('arg', 'comment', 'description', 'description_with_header',
             'equation',
             'postcondition', 'precondition',
             'self')

    def get_op_name(self):
        return '[]'

    def resolve_special(self):
        self.chk_num_args(2, None)


class DelItem(SetItem):
    def resolve_special(self):
        self.chk_num_args(1, None)


class GetItem(SetItem):
    d_sub = SetItem.d_sub + ('returns', )

    def resolve_special(self):
        self.chk_num_args(1, None)


class Condition(Description):
    d_is_def = 1
    d_sub = ('self', 'arg', 'comment', 'description', 'python_code')

    def get_arg_names(self):
        an = []
        for a in self.find_aspects('*'):
            if a.d_tag in ('self', 'arg'):
                an.append(a.src.node.arg.strip())
        return an

    def get_def_name(self):
        dn = self.src.lastname
        return dn

    def_name = property(get_def_name)


class PythonCode(Description):
    d_sub = ('comment', 'description', 'in_context')


class ConditionRef(Description):
    d_sub = ('comment', 'description',)

    def __repr__(self):
        try:
            return self.cond_expr
        except AttributeError:
            return Description.__repr__(self)

    def get_cond_id(self):
        cond_id = self.cond_definition.tgtfullname
        if self.is_not:
            cond_id = 'not ' + cond_id
        self.cond_id = cond_id
        return cond_id

    cond_id = property_nondata(get_cond_id)

    def get_definition(self):
        return self.cond_definition

    def resolve_special(self):
        cond_def = self.src.cond_definition
        self.cond_definition = self.env.get_descr_by_subject(cond_def)
        self.cond_doc_name = cond_def.parent.lastname + '.' + cond_def.lastname
        self.cond_expr = self.src.node.arg.strip()    # Mostly for information
        self.arg_names = self.src.arg_names
        self.is_not = self.src.is_not


class Precondition(ConditionRef):
    #doc_name = 'Before'
    doc_name = 'Precondition'


class Postcondition(ConditionRef):
    #doc_name = 'After'
    doc_name = 'Postcondition'


class PostcondCase:
    # Postcondition with specific variables
    def __init__(postcond, variables):
        self.postcond = postcond
        self.variables = variables


class Constructor(Description):
    d_type = 'with_args'
    d_sub = ('comment', 'description',)


class Equation(Description):
    d_sub = ('comment', 'description', 'precondition', 'postcondition')


class Args(Description):
    d_type = 'with_args'
    d_sub = ('comment', 'description', 'optionals', )

    def get_re(self, opt):
        re = self.mod.RE.Epsilon
        for a in self.find_arg_aspects():
            re += a.get_re(opt)
        return re


class NoArg(Description):
    def get_re(self, opt):
        return self.mod.RE.Epsilon


class Arg(Description):
    d_sub = ('comment', 'default', 'description', 'superkind_of', 'name', )

    def get_kind(self):
        return self.make_or_kind(self.find_kind_aspects())

    def get_name(self):
        try:
            return self.get_arg_name()
        except AttributeError:
            return '?'

    def get_arg_name(self):
        return self.src.specified_name

    def get_examples(self, get_all=False):
        examples = []
        exs = self.find_aspects('example')
        for ex in exs:
            examples.extend(ex.get_examples())
        if not exs or get_all:
            k = self.get_kind()
            examples.extend(k.get_examples())
        return examples


class KeyArgEG:
    def __init__(self, name, eg):
        self.name = name
        self.eg = eg

    def get_ex_text(self):
        return self.eg.get_ex_text()

    def get_ctx_text(self):
        return self.eg.get_ctx_text()

    def get_use_text(self, x):
        return '%s=%s' % (self.name, x)


class KeyArg(Arg):
    # Spec with keyarg means it is:
    # NOT to be used as positional argument
    # ONLY as keyword argument

    def get_examples(self, get_all=False):
        name = self.get_arg_name()
        return [KeyArgEG(name, eg) for eg in Arg.get_examples(self, get_all)]


class Draw(Description):
    d_sub = ('comment', 'description', 'key_arg', 'seq', )

    def get_re(self, opt):
        re = self.mod.RE.Epsilon
        for a in self.find_arg_aspects():
            re += a.get_re(opt)('?')
        return re


class Optionals(Description):
    d_sub = ('arg', 'args', 'key_arg', 'comment', 'seq', )
    d_type = 'with_args'

    def get_re(self, opt):
        def opt_ra(aspects):
            if not aspects:
                return self.mod.RE.Epsilon
            return (aspects[0].get_re(opt) + opt_ra(aspects[1:]))('?')
        return opt_ra(self.find_arg_aspects())


class Repeat(Description):
    d_sub = ('alt', 'arg', 'args', 'comment', 'description')

    def get_arg(self):
        return self.src.node.arg.strip()

    def get_re(self, opt):
        asp = self.find_arg_aspects()
        if not asp:
            self.error('No argument aspects.', self.src.node)

        re = asp[0].get_re(opt)
        for a in asp[1:]:
            re += a.get_re(opt)

        arg = self.get_arg()
        sep = '..'

        if sep in arg:
            args = arg.split(sep)
            if len(args) != 2:
                self.error('More than one %r in argument.' %
                           sep, self.src.node)
            lo, hi = [x.strip() for x in args]
            try:
                lo = int(lo)
            except ValueError:
                self.error('Expected int in lower bound.', self.src.node)
            if hi != '*':
                try:
                    hi = int(hi)
                except ValueError:
                    self.error('Expected int or * in upper bound.',
                               self.src.node)
        else:
            try:
                lo = int(arg)
            except ValueError:
                self.error(
                    'Expected int, int..int or int..* in argument.', self.src.node)
            hi = lo
        if lo < 0 or (hi != '*' and hi < 0):
            self.error('Expected non-negative repetition count.',
                       self.src.node)

        if hi == '*':
            res = re('*')
            for i in range(lo):
                res = re + res
        else:
            if hi < lo:
                self.error('Expected upper bound >= lower bound.',
                           self.src.node)

            a = self.mod.RE.Epsilon
            for i in range(lo):
                a += re
            b = self.mod.RE.Epsilon
            for i in range(lo, hi):
                b = (re + b)('?')
            res = a + b

        return res


class Seq(Description):
    d_sub = ('arg', 'comment', 'description', 'optionals',)
    d_sub += ('key_arg', )  # May perhaps be optionally disabled
    d_type = 'with_args'

    def get_re(self, opt):
        re = self.mod.RE.Epsilon
        for a in self.find_arg_aspects():
            re += a.get_re(opt)
        return re


class Alt(Description):
    d_sub = ('arg', 'comment', 'descripton', 'key_arg', 'no_arg', 'seq', )
    d_type = 'with_args'

    def get_re(self, opt):
        asp = self.find_arg_aspects()
        if not asp:
            self.error('No alternatives.', self.src.node)
        re = asp[0].get_re(opt)
        for a in asp[1:]:
            re |= a.get_re(opt)
        return re


class Returns(Description):
    d_sub = ('attribute', 'comment', 'description', 'description_with_header',
             'either', 'mapping', 'method')
    d_type = 'with_opt_args'

    def get_kind(self):
        return self.make_and_kind(self.find_kind_aspects())

    def get_test_kind(self):
        return self.make_and_test_kind(self.find_kind_aspects())

# help functions


def find_aspects_inseq(seq, tag):
    as_ = []
    for o in seq:
        as_.extend(o.find_aspects(tag))
    return as_

# Beam base class


class Beam:
    def __init__(self, k_tag, *objects):
        self.src = objects[0]
        self.tgt = objects[-1]
        self.k_tag = k_tag
        self.objects = objects

    def __add__(self, other):
        return compose(self, other)


class KindBeam(Beam):
    pass


class AtomKindBeam(Beam):
    pass


class KindMappingBeam(Beam):
    pass


class KindOpBeam(Beam):
    op_index = 1
    op_name_index = 1

    def find_equations(self):
        return find_aspects_inseq(self.get_op_seq(), 'equation')

    def find_postconditions(self):
        return find_aspects_inseq(self.get_op_seq(), 'postcondition')

    def find_preconditions(self):
        return find_aspects_inseq(self.get_op_seq(), 'precondition')

    def get_args_examples(self, mapname):
        top_kind = self.objects[0]
        return self.get_the_op().get_args_examples(mapname, top_kind)

    def get_op_id_name(self):
        return self.objects[self.op_name_index].get_id_name()

    def get_op_name(self):
        return self.objects[self.op_name_index].get_op_name()

    def get_op_seq(self):
        return self.objects[self.op_index:]

    def get_self_name(self):
        return self.get_the_op().get_self_name()

    def get_the_op(self):
        return self.objects[self.op_index]

    def get_return_test_kind(self):
        return self.get_the_op().get_return_test_kind()


class KindAttributeBeam(KindOpBeam):
    def get_the_op(self):
        assert 0


class KindAttributeMappingBeam(KindOpBeam):
    op_index = 2


class KindMappingBeam(KindOpBeam):
    def get_op_name(self):
        return '()'


class KOKOpBeam(KindOpBeam):
    op_index = 2
    op_name_index = 2


def subkind_of_kind(*objects):
    return beam(*objects[2:])


def compose(a, b):
    if a.tgt is not b.src:
        raise "Composition error, tgt %r is not src %r" % (a.tgt, b.src)

    objects = a.objects + b.objects[1:]
    return beam(*objects)


def remove_1_2(k_tag, *objects):
    return beam(objects[0], *objects[3:])


def remove_0(k_tag, *objects):
    return beam(*objects[1:])


beam_table = {
    ('attribute', 'attribute'): Beam,
    ('attribute', 'either'): Beam,
    ('attribute', 'kind_of'): Beam,
    ('attribute', 'kind_of', 'kind', 'attribute'): Beam,
    ('attribute', 'kind_of', 'kind', 'function_operator'): Beam,
    ('attribute', 'kind_of', 'kind', 'inplace_operator'): Beam,
    ('attribute', 'kind_of', 'kind', 'mapping'): Beam,
    ('attribute', 'kind_of', 'kind', 'operator'): Beam,
    ('attribute', 'kind_of', 'kind', 'reverse_operator'): Beam,
    ('attribute', 'kind_of', 'kind', 'delitem'): Beam,
    ('attribute', 'kind_of', 'kind', 'getitem'): Beam,
    ('attribute', 'kind_of', 'kind', 'setitem'): Beam,
    ('attribute', 'mapping'): Beam,
    ('either', ): Beam,
    ('either', 'kind'): Beam,
    ('either', 'kind', 'attribute'): Beam,
    ('kind', 'attribute'): Beam,
    ('kind', 'attribute', 'kind_of', 'kind', 'mapping'): KindAttributeBeam,
    ('kind', 'attribute', 'mapping'): KindAttributeMappingBeam,
    ('kind', 'either'): Beam,
    ('kind', 'function_operator'): KindOpBeam,
    ('kind', 'delitem'): KindOpBeam,
    ('kind', 'getitem'): KindOpBeam,
    ('kind', 'inplace_operator'): KindOpBeam,
    ('kind', 'kind_of'): Beam,
    ('kind', 'kind_of', 'kind', 'attribute'): Beam,
    ('kind', 'mapping'): KindMappingBeam,
    ('kind', 'operator'): KindOpBeam,
    ('kind', 'reverse_operator'): KindOpBeam,
    ('kind', 'setitem'): KindOpBeam,
    ('kind', 'subkind_of'): Beam,
    ('kind', 'subkind_of', 'kind', 'attribute'): remove_1_2,
    ('kind', 'subkind_of', 'kind', 'function_operator'): remove_1_2,
    ('kind', 'subkind_of', 'kind', 'delitem'): remove_1_2,
    ('kind', 'subkind_of', 'kind', 'getitem'): remove_1_2,
    ('kind', 'subkind_of', 'kind', 'inplace_operator'): remove_1_2,
    ('kind', 'subkind_of', 'kind', 'mapping'): remove_1_2,
    ('kind', 'subkind_of', 'kind', 'operator'): remove_1_2,
    ('kind', 'subkind_of', 'kind', 'reverse_operator'): remove_1_2,
    ('kind', 'subkind_of', 'kind', 'setitem'): remove_1_2,
    ('kind_of', 'kind'): Beam,
    ('kind_of', 'kind', 'attribute'): Beam,
    ('kind_of', 'kind', 'function_operator'): KOKOpBeam,
    ('kind_of', 'kind', 'delitem'): KOKOpBeam,
    ('kind_of', 'kind', 'getitem'): KOKOpBeam,
    ('kind_of', 'kind', 'inplace_operator'): KOKOpBeam,
    ('kind_of', 'kind', 'operator'): KOKOpBeam,
    ('kind_of', 'kind', 'reverse_operator'): KOKOpBeam,
    ('kind_of', 'kind', 'setitem'): KOKOpBeam,
    ('kind_of', 'kind', 'mapping'): Beam,
    ('subkind_of', 'kind'): Beam,
    ('subkind_of', 'kind', 'attribute'): Beam,
    ('subkind_of', 'kind', 'function_operator'): Beam,
    ('subkind_of', 'kind', 'delitem'): Beam,
    ('subkind_of', 'kind', 'getitem'): Beam,
    ('subkind_of', 'kind', 'inplace_operator'): Beam,
    ('subkind_of', 'kind', 'mapping'): Beam,
    ('subkind_of', 'kind', 'operator'): Beam,
    ('subkind_of', 'kind', 'reverse_operator'): Beam,
    ('subkind_of', 'kind', 'setitem'): Beam,

}


def beam(*objects):
    k_tag = tuple([x.d_tag for x in objects])
    C = beam_table[k_tag]
    return C(k_tag, *objects)


class ProductSubject:
    def __init__(self, subjects):
        self.subjects = subjects
        self.fullname = '(%s)' % '*'.join([x.fullname for x in subjects])


class Subject:
    args = ()
    specified_name = None

    def __init__(self, parent, node, lastname):
        self.parent = parent
        self.pac = parent.pac
        self.mod = self.pac.mod
        self.node = node
        self.filename = self.pac.filename
        self.lastname = lastname
        self.aspects = []
        self.subjects = {}
        self.node_index = 0
        self.tag = node.tag
        self.description_class = self.mod.get_description_class(node.tag)
        self.aspect_mode = None

        if self.parent is not self:
            self.fullname = self.parent.make_child_name(self.lastname)
        else:
            self.fullname = self.lastname

    def _visit_type_definition(self, node):
        names = self.get_arglist(node, min=1)
        for name in names:
            self.add_new_subject(node, name)

    def _visit_type_operator(self, node):
        shtag = self.mod.SpecNodes.reverse_node_aliases[node.tag]
        names = self.get_arglist(node, min=1)
        for name in names:
            name = '%s:%s' % (shtag, name)
            self.add_new_subject(node, name)

    def _visit_type_other(self, node):
        self.add_new_subject(node)

    def _visit_type_with_args(self, node):
        names = self.get_arglist(node)
        args = [self.find_subject(name, node) for name in names]
        subject = self.add_new_subject(node)
        if args:
            subject.args = args

    def _visit_type_with_opt_args(self, node):
        names = self.get_arglist(node, min=0)
        args = [self.find_subject(name, node) for name in names]
        subject = self.add_new_subject(node)
        if args:
            subject.args = args

    def add_new_subject(self, node, lastname=None):
        subject = self.new_subject(node, lastname)
        self.add_subject(subject)
        return subject

    def add_subject(self, subject):
        self.def_subject(subject)
        subject.add_top_node()
        return subject

    def add_top_node(self):
        node = self.node
        self._visit_children(node)

    def def_new_subject(self, node, lastname=None):
        subject = self.new_subject(node, lastname)
        self.def_subject(subject)
        return subject

    def def_subject(self, subject):
        if subject.description_class.d_is_def:
            name = subject.lastname
            if name in self.subjects:
                self.error('Redefinition of %r.' % name, subject.node,
                           more=[(
                               'Previous definition of %r.' % name,
                               self.subjects[name].node)]
                           )
                return  # For clarity; there's most certainly an exception

            subject.definame = name
            self.subjects[name] = subject
        else:
            subject.definame = None
        self.aspects.append(subject)

    def error(self, msg, node=None, exception=ReportedError, **kwds):
        return self.pac.error(msg, node, exception, **kwds)

    def find_subject(self, name, node):
        return self.pac.find_subject(name, node, self)

    def get_arglist(self, node, min=0):
        arglist = node.get_arglist()
        for arg in node.get_arglist():
            if not arg:
                if node.arg.strip().startswith(',') or node.arg.strip().endswith(','):
                    m = 'Arg list to definition can not start or end with a comma.'
                else:
                    m = 'Missing argument to definition.'
                self.error(m, node, exception=None)
                arglist = [x for x in arglist if x]
                break
        if len(arglist) < min:
            self.error(
                'Not enough arguments, minimum %d expected to node %s' % (
                    min, node),
                node)
        return arglist

    def get_arglist_only(self, node, min=0):
        al = self.get_arglist(node, min)
        self.no_children(node)
        return al

    def get_line(self, index):
        try:
            with open(self.filename) as f:
                text = list(f.readlines())[index].rstrip()
        except Exception:
            text = None
        return text

    def _visit_aspect(self, node, mode):
        if self.aspect_mode is None:
            self.aspect_mode = mode
        else:
            if self.aspect_mode != mode:
                self.error('Inconsistent aspect mode: %r, was: %r' % (mode, self.aspect_mode),
                           node)
        self._visit_children(node)

    def _visit_children(self, node):
        for ch in node.children:
            try:
                if ch.tag not in self.description_class.d_sub:
                    self.error('Invalid  tag: %r  in: %r. Allowed = %s' % (
                        ch.tag, self.tag, self.description_class.d_sub), node)
                if self.mod.cover_check is not None:
                    self.mod.cover_check.setdefault(self.tag, {})[ch.tag] = 1
                ch.accept(self)
            except ReportedError:
                pass
            self.node_index += 1

    def make_child_name(self, child_lastname):
        return '%s.%s' % (self.fullname, child_lastname)

    def new_subject(self, node, name=None):
        is_def = self.mod.get_description_class(node.tag).d_is_def
        assert is_def == (name is not None)
        if name is None:
            name = '<%d>' % self.node_index
        tag = node.tag
        if tag == 'macro':
            return MacroSubject(self, node, name)
        elif tag == 'document':
            return DocumentSubject(self, node, name)
        else:
            return Subject(self, node, name)

    def new_tag_node(self, tag, node):
        return self.mod.SpecNodes.node_of_taci(tag, '', node.children, node.index)

    def no_children(self, node):
        if node.children:
            self.error('No children expected for node with tag %r' % node.tag,
                       node,
                       exception=None)

    def visit_and(self, node):
        for name in self.get_arglist(node, min=1):
            ofsubject = self.find_subject(name, node)
            ofsubject._visit_aspect(node, 'and')

    def visit_aspects_of(self, node):
        for name in self.get_arglist(node, min=1):
            ofsubject = self.find_subject(name, node)
            ofsubject._visit_aspect(node, 'aspect')

    def visit_arg(self, node, must_have_name=False):
        arg = node.arg.strip()
        arg_name = None
        kind = None
        if arg:
            if ':' in arg:
                nk = arg.split(':')
                if len(nk) > 2:
                    self.error('More than 1 colon in argument.', node)
                name, kind_name = [x.strip() for x in nk]
                if kind_name:
                    kind = self.find_subject(kind_name, node)
                if name:
                    arg_name = name
            else:
                # Is there an obvious default ?
                # For KeyArg, yes, the name is always.
                # let's say it's the name
                arg_name = arg

        subject = self.new_subject(node)
        if arg_name:
            subject.specified_name = arg_name
        self.add_subject(subject)

        if must_have_name and subject.specified_name is None:
            self.error('No argument name specified.', node)

        if kind is not None:
            subject.args = [kind]

    def visit_comment(self, node):
        pass

    def visit_condition(self, node):
        names = self.get_arglist(node, min=1)
        for name in names:
            self.add_new_subject(node, 'cond:%s' % name)

    def visit_default(self, node):
        description_class = self.mod.get_description_class(node.tag)
        arg = node.arg.strip()
        colon = arg.startswith(':')
        if (description_class.d_type == 'definition') != colon:
            if colon:
                msg = 'Tag %r is not a definition, should not have ::' % node.tag
            else:
                msg = 'Tag %r is a definition, requires ::' % node.tag
            self.error(msg, node, exception=None)

        getattr(self, '_visit_type_%s' % description_class.d_type)(node)

    def visit_description(self, node):
        self.def_new_subject(node)

    def visit_description_with_header(self, node):
        self.visit_description(node)

    def visit_example(self, node):
        subject = self.add_new_subject(node)
        partab = subject.description_class.partab
        ex = node.arg.strip()

        if '\n' in ex:
            if not (partab.get(ex[:1]) == ex[-1:] or
                    partab.get(ex[:3]) == ex[-3:]):
                self.error('Multi-line expression should be in parentheses (for clarity).', node,
                           exception=None, harmless=1)
            ex = '(%s)' % ex

        subject.ex_text = ex

    def visit_import(self, node):

        my_names = self.get_arglist(node, min=1)
        resolve_mode = None
        usings = None
        froms = []
        for ch in node.children:
            t = ch.tag
            if t == 'from':
                for name in self.get_arglist_only(ch):
                    froms.append((name, ch))
            elif t == 'resolve_by':
                if resolve_mode:
                    self.error("More than 1 'resolve' clause.",
                               ch.node, exception=None)
                else:
                    resolve_mode = ch.arg.strip()
                    if not resolve_mode in ('and', 'or'):
                        self.error("Resolve by: and / or expected.",
                                   ch,
                                   exception=None)
                        resolve_mode = 'and'
            elif t == 'using':
                if usings is None:
                    usings = []
                for name in self.get_arglist_only(ch):
                    usings.append((name, ch))
            else:
                self.error('Unexpected clause in import', ch, exception=None)

        using_map = {}
        if usings is not None:
            if len(usings) != len(my_names):
                if len(using_names) < len(my_names):
                    manyfew = 'few'
                else:
                    manyfew = 'many'
                self.error(
                    "Too %s 'using' names, should match number of names in .import" % manyfew,
                    using_node,
                    exception=None)
            for m, u in zip(my_names, usings):
                # zip stops at the shortest list, ok
                using_map[m] = u

        if len(froms) == 0:
            self.error("No 'from' clause", node)

        if len(froms) > 1:
            if not resolve_mode:
                self.error("Importing from multiple packages but no 'resolve by' clause",
                           node, exception=None)
                resolve_mode = 'and'

        for name in my_names:
            subject = self.def_new_subject(node, name)
            subject.imp_resolve_mode = resolve_mode
            subject.imp_using_map = using_map
            subject.imp_froms = froms

    def visit_key_arg(self, node):
        self.visit_arg(node, must_have_name=True)

    def visit_method(self, node):
        arg = node.arg.strip()
        if not arg.startswith(':'):
            self.error("Tag 'method' is a definition, requires ::", node)
        self.mod.node_of_taci('attribute', arg,
                              (self.mod.node_of_taci('mapping', '', node.children),)).accept(self)

    def visit_name(self, node):
        if self.specified_name is not None:
            self.error('Duplicate name specification.', node)
        name = node.arg.strip()
        if not name:
            self.error('No name specification.', node)
        self.specified_name = name

    def visit_or(self, node):
        for name in self.get_arglist(node, min=1):
            ofsubject = self.find_subject(name, node)
            ofsubject._visit_aspect(node, 'or')

    def visit_postcondition(self, node):
        arg = node.arg.strip()
        if not '(' in arg:
            self.error('No left parenthesis', node)
        lpar = arg.index('(')
        rpar = arg.find(')')
        if rpar < lpar:
            self.error('None or misplaced right parenthesis', node)

        n = arg[lpar+1:rpar].strip()
        if ',' in n:
            n = [x.strip() for x in n.split(',')]
        else:
            n = [n]
        arg_names = n

        cond_name = arg[:lpar].strip()
        if not cond_name:
            self.error('No condition name', node)
        is_not = 0
        if cond_name.startswith('not '):
            cond_name = cond_name[4:].strip()
            is_not = 1

        parts = cond_name.split('.')
        if not parts[-1].startswith('cond:'):
            parts[-1] = 'cond:'+parts[-1]
            cond_name = '.'.join(parts)

        cond_def = self.find_subject(cond_name, node)
        subject = self.add_new_subject(node)
        subject.cond_definition = cond_def
        subject.cond_name = cond_name
        subject.arg_names = arg_names
        subject.is_not = is_not

    def visit_precondition(self, node):
        self.visit_postcondition(node)


class ErrorSubject(Subject):
    pass


class PackageSubject(Subject):
    def __init__(self, mod, specenv, node, name, filename):
        self.mod = mod
        self.specenv = specenv
        self.pac = self
        self.filename = filename
        #name = 'package_%s'%(name,)
        name = '%s' % (name,)
        Subject.__init__(self, self, node, name)
        self.lastname = name.split('.')[-1]
        self.tag = 'package'
        self.description_class = Package

        self.documents = []
        for s in mod.predefined_subjects:
            s = s(self)
            self.subjects[s.fullname] = s

        self._visit_children(node)
        del self.specenv  # It was used only for error report

    def error(self, msg, node=None, exception=ReportedError, **kwds):
        return self.specenv.error(msg, node, exception, **kwds)

    def find_subject(self, name, node, context=None):
        if not name:
            self.error('Invalid subject name: %r' % name, node)

        parts = [x.strip() for x in name.split('.')]
        if not parts[0]:
            tag = parts[1]
            parts = parts[2:]
        else:
            tag = 'myfile'

        if tag == 'myfile':
            s = self
        elif tag == 'mykind':
            s = context
            if s is not None:
                kind_tags = ('kind', 'and', 'import')
                while s.parent != self and s.tag not in kind_tags:
                    s = s.parent
                if s.tag not in kind_tags:
                    s = None
            if s is None:
                self.error('mykind tag without such a context: %r' %
                           name, node)
        else:
            self.error('Invalid tag %r in %r' % (tag, name), node)

        sname = s.lastname
        for i, n in enumerate(parts):
            ns = s.subjects.get(n)
            if ns is None:
                if s.tag != 'import':
                    self.error('No such subject: %r  in %r.' %
                               (n, sname), node)
                return SubImportSubject(s, node, parts[i:])
            sname = sname + '.' + n
            s = ns
        return s


class SubImportSubject:
    def __init__(self, parent, node, rnparts):
        self.parent = parent
        self.node = node
        self.rnparts = rnparts
        self.fullname = '.'.join([parent.fullname]+rnparts)
        self.lastname = rnparts[-1]


class MacroSubject(Subject):
    def add_top_node(self):
        pass


class DocumentSubject(Subject):
    def add_top_node(self):
        self.parent.documents.append(self)


class GuppyWorld(Subject):
    def __init__(self, env):
        self.pac = env
        self.fullname = self.lastname = "Guppy_World"
        self.node = None
        self.tag = '<GuppyWorld>'
        self.aspects = []
        self.description_class = Description


class _GLUECLAMP_:
    _imports_ = (
        '_parent:Document',
        '_parent:FileIO',
        '_parent.FileIO:IO',
        '_parent:Filer',
        '_parent:Html',
        '_parent:Latex',
        '_parent:SpecNodes',
        '_parent.SpecNodes:node_of_taci',
        '_parent:Tester',
        '_root.hashlib:md5',
        '_root.guppy.etc:iterpermute',
        '_root.guppy.etc:RE',
    )

    _chgable_ = ('cover_check', 'io_dir', 'max_errors')

    description_classes = {
        'alt': Alt,
        'arg': Arg,
        'args': Args,
        'attribute': Attribute,
        'comment': Comment,
        'condition': Condition,
        'constructor': Constructor,
        'default': Default,
        'defines': Defines,
        'delitem': DelItem,
        'description': DescriptionDescription,
        'description_with_header': DescriptionWithHeader,
        'equation': Equation,
        'example': Example,
        'either': Either,
        'draw': Draw,
        'function_operator': FunctionOperator,
        'getitem': GetItem,
        'import': Import,
        'in_context': InContext,
        'inplace_operator': InplaceOperator,
        'key_arg': KeyArg,
        'kind': Kind,
        'kind_of': KindOf,
        'macro': Macro,
        'mapping': Mapping,
        'no_arg': NoArg,
        'operator': Operator,
        'postcondition': Postcondition,
        'precondition': Precondition,
        'python_code': PythonCode,
        'reverse_operator': ReverseOperator,
        'optionals': Optionals,
        'package': Package,
        'repeat': Repeat,
        'returns': Returns,
        'self': Self,
        'seq': Seq,
        'setitem': SetItem,
        'subkind_of': SubkindOf,
        'superkind': Superkind,
        'superkind_of': SuperkindOf,
    }

    tgt_prefix = '.tgt.'

    cover_check = None
    io_dir = None
    max_errors = 10

    def get_description_class(self, tag):
        return self.description_classes.get(tag, Description)

    def _get_predefined_subjects(self):
        return (GuppyWorld,)

    def _get_package_cache(self):
        return {}

    def main(self, filename, **kwds):
        se = SpecEnv(self)
        se.process_main(filename, **kwds)

    def _test_main_(self):
        pass

    def set_input_dir(self, dir):
        dir = self.IO.path.abspath(dir)
        self.input_dir = dir



# Constants used in communication between RemoteAccess and Monitor

READLINE = 'READLINE\n'
KEYBOARDINTERRUPT = 'KEYBOARDINTERRUPT\n'
DONE = 'DONE\n'
LOCALHOST = '127.0.0.1'
#HEAPYPORT = 8834
HEAPYPORT = 3546



# Tests of equivalence relations.

# These are also tested by test_Classifiers.
# This is some more tests, tailored esp. to the user view.
# (test_Classifiers was so slow already, so I start over)

# o Intended to be exhaustive wrt all ER's defined
#
# o Intersection of ER's

from guppy.heapy.test import support
import inspect
import os.path


class TestCase(support.TestCase):
    pass


class FirstCase(TestCase):
    def test_1(self):
        hp = self.heapy.Use
        hp.reprefix = 'hp.'

        a = hp.iso(1, '', 'asdf', 3.4, 3.7, 2)

        ts = (hp.Type & hp.Size)
        k = ts[a]

        # From Sep 1-2 2005
        # (h&dict).by(hp.Id.dictof&hp.Size)._get_partition()
        # (h&dict).by((hp.Type&hp.Size).dictof&hp.Size)

        # These require with_referrers of refby/via classifier
        # after gc collect referrers graph will be empty
        # (h).by(hp.Module.refby.dictof)
        # (h).by(hp.Via.dictof)

        # How to construct RCS / refby

        # self.aseq(hp.Type.refby(int, list) , hp.Type.refby(list, int)

        class C:
            pass

        di = hp.iso(C.__dict__, [])

        # db = di.by('Rcs')
        # for i in (0, 1):
        #     rk = repr(db[i].kind)
        #     ek = eval(rk, {'hp': hp}, {})
        #     self.aseq(ek, db[i].kind)
        #     self.aseq(db & ek, db[i])

    def test_2(self):
        ' Systematically test all kind constructors: '
        # wrt repr and evaluation of repr

        hp = self.heapy.Use
        hp.reprefix = 'hp.'

        class C:
            pass

        class T(object):
            pass

        c = C()
        t = T()
        import sys

        for s in (
            'hp.Clodo(dictof=C)',
            'hp.Clodo(dictof=T)',
            'hp.Clodo(dictof=())',
            'hp.Clodo(C)',
            'hp.Clodo(T)',
            'hp.Id(id(c))',
            'hp.Module("sys")',

            'hp.Prod()',
            'hp.Prod("<string>")',
            'hp.Prod(T)',
            'hp.Prod("/foo.py", 1)',
            'hp.Prod("/foo.py", None)',
            'hp.Prod(None, 1)',

            'hp.Rcs(hp.Clodo.sokind(int)(dictof=C))',
            'hp.Size(hp.iso(c).indisize)',
            'hp.Size(hp.iso(C).indisize).dictof',

            'hp.Type(T)',
            'hp.Type(int)',

            'hp.Unity()',

            'hp.Via()',
            # Via is also specially tested below
        ):
            x = eval(s, {'hp': hp, 'C': C, 'T': T, 'c': c}, {})
            rx = repr(x)
            self.aseq(eval(rx, {'hp': hp}, {}), x)

        for i, s in enumerate((
            # Test Via construction.
            # One test for each relation kind defined in Path except IDENTITY and RELSRC.
            # In code order.
            "hp.Via('_.x')",
            "hp.Via('_[0]')",
            "hp.Via('_.keys()[0]')",
            "hp.Via('_->abc')",
            "hp.Via('_.__dict__.keys()[0]')",
            "hp.Via('_.f_locals[\"abc\"]')",
            "hp.Via('_.f_locals [\"abc\"]')",
            "hp.Via('_->f_valuestack[0]')",
            "hp.Via('list(_)[0]')",
        )):
            code = i + 1
            x = eval(s, {'hp': hp}, {})
            rel = list(x.arg)[0]
            self.aseq(rel.kind, code)
            rx = repr(x)
            self.aseq(eval(rx, {'hp': hp}, {}), x)

    def test_3(self):
        ' Test of dictof '
        # Test of dictof on something that requires memoization, i.e. Size, & (and)

        hp = self.heapy.Use

        class C:
            pass

        class T(object):
            # The test works only if sizes of objects of class C and T differ.
            # At first test, T() was 4 bytes smaller than C().
            # This might be brittle with different systems.
            # This is to make sure this diff gets significantly bigger:
            __slots__ = '__dict__', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'

        c = C()
        t = T()
        dn = {}
        isod = hp.iso(c.__dict__, t.__dict__, dn)
        for x in (
                t, c):
            X = x.__class__
            for k in (
                hp.Clodo(dictof=X),
                hp.Size(hp.iso(x).indisize).dictof,
                hp.iso(x).bysize.kind.dictof,
                hp.iso(x.__dict__).kind,
            ):
                self.aseq(isod & k,  hp.iso(x.__dict__))

        # Test no-owner selection
        for k in (
            hp.Nothing.dictof,
        ):
            self.aseq(isod & k, hp.iso(dn))

    def test_4(self):
        ' Test of via '
        # Esp. representation, construction

        class C:
            pass

        c = C()
        hp = self.heapy.Use

        isod = hp.iso(c.__dict__)

        x = isod.by('Via').kind
        self.aseq(repr(x), "hpy().Via('.__dict__')")

    def test_5(self):
        ' Non-systematic tests that came up around Sep 14 2005 '

        class C:
            pass

        c = C()
        d = {}
        cref = [c]
        cref.append(cref)
        c.cref = cref
        hp = self.heapy.Use
        hp.reprefix = 'hp.'

        # I thought these should be the same
        a = hp.iso(C, c, c.__dict__, d)
        b = hp.iso(C, c, c.__dict__, d)
        self.aseq(a, b)

        # This is a kind of nested refdby that has been a concern lately
        # -- how to represent it

        s = hp.iso(C.__dict__, C, c, c.__dict__,
                   d).by(hp.Clodo.refdby.refdby)

        for i in range(len(s)):
            a = s[i].kind
            ra = repr(a)
            era = eval(ra, {'hp': hp}, {})
            self.aseq(a, era)

            self.aseq(s & era,
                      s[i])

        import sys

        p = sys.path
        del sys

        s = hp.iso(p)
        x = s.by(hp.Module.dictof.refdby)
        self.aseq(s & eval(repr(x.kind), {'hp': hp}, {}), s)

    def test_6(self):
        ' Test of .refdby on all others '
        with self.tracemalloc_state():
            class C:
                pass

            c = C()
            d = {}
            cref = [c]
            cref.append(cref)
            c.cref = cref

            hp = self.heapy.Use
            hp.reprefix = 'hp.'

            import sys
            s = hp.iso(C.__dict__, C, c, c.__dict__, d, sys)

            for pre in (
                'Unity',
                'Clodo',
                'Id',
                'Module',
                'Prod',
                ('Rcs', 0),
                'Size',
                'Type',
                'Via'
            ):
                if isinstance(pre, tuple):
                    pre, level = pre
                else:
                    level = 1
                er = getattr(hp, pre)
                self.er_test(er, s, level)

    def er_test(self, er, set, level=1):
        # Tests what any eqv. rel. er should do
        hp = self.heapy.Use

        rer = repr(er)
        self.aseq(eval(rer, {'hp': hp}, {}), er)

        for s in (set,):
            sby = s.by(er)
            sk = sby.kind
            rsk = repr(sk)
            ske = eval(rsk, {'hp': hp}, {})

            self.aseq(ske, sk, -1)
            self.aseq(s & sk, s, -1)
            self.aseq(s & ske, s, -1)

        # That it can do .refdby

        er_refdby = er.refdby

        # That it can do .dictof

        er_dictof = er.dictof

        if level > 0:
            self.er_test(er_refdby, set, level - 1)
            self.er_test(er_dictof, set, level - 1)

    def test_7(self):
        ' Test of alternative sets w. biper '

        hp = self.heapy.Use

        class C:
            pass

        class D(C):
            pass

        class E(D):
            pass

        class T(object):
            pass

        class U(T):
            pass

        class V(U):
            pass

        c = C()
        d = D()
        e = E()
        t = T()
        u = U()
        v = V()

        s = hp.iso([], {}, c, d, e, t, u, v, d.__dict__)

        for k in (
            hp.Size(32),
            hp.Type(U),
        ):

            lt = k.alt('<')
            le = k.alt('<=')
            ge = k.alt('>=')
            gt = k.alt('>=')
            ne = k.alt('!=')
            assert (s & le) & (s & ge) == s & k

            for a in (lt, le, ge, gt, ne, le & ~k):
                s & a

        import gc

        gc.collect()

        a = s.by(k.alt('>=').biper)
        b = s.by(k.alt('>=').biper)
        self.assertTrue(hp.iso(d.__dict__) <= a[0])
        self.assertTrue(a == b)

        gc.collect()

        a = s.by(k.alt('<=').biper)
        b = s.by(k.alt('<=').biper)
        self.assertTrue(hp.iso(d.__dict__) <= a[0])
        self.assertTrue(a == b)

    def test_8(self):
        ' Test of findex and biper '
        # added Nov 3 2005

        hp = self.heapy.Use

        class C:
            pass
        c = C()
        li = []
        di = {}

        s = hp.iso(li, di, c, 1)

        for k, i in (
            (hp.Type(dict), 0),
            (hp.iso(c), 1),
            (hp.iso(c, li), 1),
        ):
            p = s.by(k.biper)
            self.aseq(p[i].kind.fam.classifier.kinds[0], k)

    def test_9(self):
        ' Test the subrelation relation '

        from guppy import hpy
        ernames = ['Clodo', 'Id', 'Idset', 'Module',
                   'Prod', 'Rcs', 'Size', 'Type',
                   'Unity']

        hp = hpy()

        ers = [(name, getattr(hp, name)) for name in ernames]
        ers.append(('Size&Type', hp.Size & hp.Type))

        from io import StringIO
        f = StringIO()
        print(''.ljust(10), end=' ', file=f)
        for b in ers:
            print(b[0].ljust(7), end=' ', file=f)
        print(file=f)

        for a in ers:
            print(a[0].ljust(10), end=' ', file=f)
            for b in ers:
                print(str((a[1] < b[1]))[:1].ljust(7), end=' ', file=f)
            print(file=f)
        self.aseq(f.getvalue(), """\
           Clodo   Id      Idset   Module  Prod    Rcs     Size    Type    Unity   Size&Type \n\
Clodo      F       F       F       F       F       F       F       T       T       F       \n\
Id         F       F       F       F       F       F       F       F       T       F       \n\
Idset      F       F       F       F       F       F       F       F       T       F       \n\
Module     F       F       F       F       F       F       F       F       T       F       \n\
Prod       F       F       F       F       F       F       F       F       T       F       \n\
Rcs        F       F       F       F       F       F       F       F       T       F       \n\
Size       F       F       F       F       F       F       F       F       T       F       \n\
Type       F       F       F       F       F       F       F       F       T       F       \n\
Unity      F       F       F       F       F       F       F       F       F       F       \n\
Size&Type  F       F       F       F       F       F       T       T       T       F       \n\
""")

    def test_10(self):
        ' Test of producer profile '
        hp = self.heapy.Use

        with self.tracemalloc_state():
            class C:
                pass

            file = inspect.currentframe().f_code.co_filename

            def lineno():
                return inspect.currentframe().f_back.f_lineno

            str_aloc = eval('C()', {'C': C}, {})
            test_aloc, test_aloc_line = C(), lineno()

            def subf():
                return C(), lineno()

            subf_aloc, subf_aloc_line = subf()

            isod = hp.iso(None, str_aloc, test_aloc, subf_aloc)

            self.aseq(isod & hp.Prod(), hp.iso(None))
            self.aseq(isod & hp.Prod('<string>', 1), hp.iso(str_aloc))
            self.aseq(isod & hp.Prod(file, test_aloc_line), hp.iso(test_aloc))
            self.aseq(isod & hp.Prod(file, subf_aloc_line), hp.iso(subf_aloc))

            self.aseq(isod & hp.Prod(file), hp.iso(test_aloc, subf_aloc))
            self.aseq(isod & hp.Prod(os.path.dirname(file)),
                      hp.iso(test_aloc, subf_aloc))

            self.aseq(isod & hp.Prod(self.test_10), hp.iso(test_aloc, subf_aloc))
            self.aseq(isod & hp.Prod(subf), hp.iso(subf_aloc))

            self.aseq(isod & hp.Prod(None, test_aloc_line).alt('>=') &
                      hp.Prod(None, test_aloc_line).alt('<='),
                      hp.iso(test_aloc))

        with self.assertRaises(TypeError):
            hp.Prod('', lineno=1)

        with self.assertRaises(TypeError):
            hp.Prod(filename='', line=1)

        with self.assertRaises(TypeError):
            hp.Prod(None)

        with self.assertRaises(TypeError):
            hp.Prod(1, 1)

        with self.assertRaises(TypeError):
            hp.Prod('', '')

        with self.assertRaises(TypeError):
            hp.Prod('', 1, 2)


def test_main(debug=0):
    support.run_unittest(FirstCase, debug)


if __name__ == "__main__":
    from guppy.heapy.Remote import off
    off()

    test_main()
